import pytest

from .fixtures import clear_registry
from ramses import registry


@pytest.mark.usefixtures('clear_registry')
class TestRegistry(object):

    def test_add_decorator(self):
        @registry.add
        def foo(*args, **kwargs):
            return args, kwargs

        assert registry.registry['foo'] is foo
        assert list(registry.registry.keys()) == ['foo']

    def test_add_decorator_with_name(self):
        @registry.add('bar')
        def foo(*args, **kwargs):
            return args, kwargs

        assert registry.registry['bar'] is foo
        assert list(registry.registry.keys()) == ['bar']

    def test_add_arbitrary_object(self):
        registry.add('foo', 1)
        registry.add('bar', 2)

        assert registry.registry['foo'] == 1
        assert registry.registry['bar'] == 2
        assert sorted(registry.registry.keys()) == ['bar', 'foo']

    def test_get(self):
        registry.registry['foo'] = 1
        assert registry.get('foo') == 1

    def test_get_error(self):
        assert not list(registry.registry.keys())
        with pytest.raises(KeyError) as ex:
            registry.get('foo')
        assert 'is not registered in ramses registry' in str(ex.value)

    def test_mget(self):
        registry.registry['Foo.bar'] = 1
        registry.registry['Foo.zoo'] = 2
        assert registry.mget('FoO') == {'bar': 1, 'zoo': 2}

    def test_mget_not_existing(self):
        registry.registry['Foo.bar'] = 1
        registry.registry['Foo.zoo'] = 2
        assert registry.mget('asdasdasd') == {}



import pytest
from mock import Mock, patch, call

from ramses import generators
from .fixtures import engine_mock


class TestHelperFunctions(object):
    @patch.object(generators, 'get_static_parent')
    def test_get_nefertari_parent_resource_no_parent(self, mock_get):
        mock_get.return_value = None
        assert generators._get_nefertari_parent_resource(1, 2, 3) == 3
        mock_get.assert_called_once_with(1)

    @patch.object(generators, 'get_static_parent')
    def test_get_nefertari_parent_resource_parent_not_defined(
            self, mock_get):
        mock_get.return_value = Mock(path='foo')
        assert generators._get_nefertari_parent_resource(
            1, {}, 3) == 3
        mock_get.assert_called_once_with(1)

    @patch.object(generators, 'get_static_parent')
    def test_get_nefertari_parent_resource_parent_defined(
            self, mock_get):
        mock_get.return_value = Mock(path='foo')
        assert generators._get_nefertari_parent_resource(
            1, {'foo': 'bar'}, 3) == 'bar'
        mock_get.assert_called_once_with(1)

    @patch.object(generators, 'generate_resource')
    def test_generate_server_no_resources(self, mock_gen):
        generators.generate_server(Mock(resources=None), 'foo')
        assert not mock_gen.called

    @patch.object(generators, '_get_nefertari_parent_resource')
    @patch.object(generators, 'generate_resource')
    def test_generate_server_resources_generated(
            self, mock_gen, mock_get):
        config = Mock()
        resources = [
            Mock(path='/foo'),
            Mock(path='/bar'),
        ]
        generators.generate_server(Mock(resources=resources), config)
        assert mock_get.call_count == 2
        mock_gen.assert_has_calls([
            call(resources[0], mock_get()),
            call(resources[1], mock_get()),
        ])

    @patch.object(generators, '_get_nefertari_parent_resource')
    @patch.object(generators, 'generate_resource')
    def test_generate_server_call_per_path(
            self, mock_gen, mock_get):
        config = Mock()
        resources = [
            Mock(path='/foo'),
            Mock(path='/foo'),
        ]
        generators.generate_server(Mock(resources=resources), config)
        assert mock_get.call_count == 1
        mock_gen.assert_called_once_with(resources[0], mock_get())


@pytest.mark.usefixtures('engine_mock')
class TestGenerateModels(object):

    @patch('ramses.generators.is_dynamic_uri')
    def test_no_resources(self, mock_dyn):
        generators.generate_models(config=1, raml_resources=[])
        assert not mock_dyn.called

    @patch('ramses.models.handle_model_generation')
    def test_dynamic_uri(self, mock_handle):
        generators.generate_models(
            config=1, raml_resources=[Mock(path='/{id}')])
        assert not mock_handle.called

    @patch('ramses.models.handle_model_generation')
    def test_no_post_resources(self, mock_handle):
        generators.generate_models(config=1, raml_resources=[
            Mock(path='/stories', method='get'),
            Mock(path='/stories', method='options'),
            Mock(path='/stories', method='patch'),
        ])
        assert not mock_handle.called

    @patch('ramses.generators.attr_subresource')
    @patch('ramses.models.handle_model_generation')
    def test_attr_subresource(self, mock_handle, mock_attr):
        mock_attr.return_value = True
        resource = Mock(path='/stories', method='POST')
        generators.generate_models(config=1, raml_resources=[resource])
        assert not mock_handle.called
        mock_attr.assert_called_once_with(resource, 'stories')

    @patch('ramses.generators.attr_subresource')
    @patch('ramses.models.handle_model_generation')
    def test_non_auth_model(self, mock_handle, mock_attr):
        mock_attr.return_value = False
        mock_handle.return_value = ('Foo', False)
        config = Mock()
        resource = Mock(path='/stories', method='POST')
        generators.generate_models(
            config=config, raml_resources=[resource])
        mock_attr.assert_called_once_with(resource, 'stories')
        mock_handle.assert_called_once_with(resource, 'stories')
        assert config.registry.auth_model != 'Foo'

    @patch('ramses.generators.attr_subresource')
    @patch('ramses.models.handle_model_generation')
    def test_auth_model(self, mock_handle, mock_attr):
        mock_attr.return_value = False
        mock_handle.return_value = ('Foo', True)
        config = Mock()
        resource = Mock(path='/stories', method='POST')
        generators.generate_models(
            config=config, raml_resources=[resource])
        mock_attr.assert_called_once_with(resource, 'stories')
        mock_handle.assert_called_once_with(resource, 'stories')
        assert config.registry.auth_model == 'Foo'


class TestGenerateResource(object):
    def test_dynamic_root_parent(self):
        raml_resource = Mock(path='/foobar/{id}')
        parent_resource = Mock(is_root=True)
        with pytest.raises(Exception) as ex:
            generators.generate_resource(raml_resource, parent_resource)

        expected = ("Top-level resources can't be dynamic and must "
                    "represent collections instead")
        assert str(ex.value) == expected

    def test_dynamic_not_root_parent(self):
        raml_resource = Mock(path='/foobar/{id}')
        parent_resource = Mock(is_root=False)
        new_resource = generators.generate_resource(
            raml_resource, parent_resource)
        assert new_resource is None

    @patch('ramses.generators.dynamic_part_name')
    @patch('ramses.generators.singular_subresource')
    @patch('ramses.generators.attr_subresource')
    @patch('ramses.models.get_existing_model')
    @patch('ramses.generators.generate_acl')
    @patch('ramses.generators.resource_view_attrs')
    @patch('ramses.generators.generate_rest_view')
    def test_full_run(
            self, generate_view, view_attrs, generate_acl, get_model,
            attr_res, singular_res, mock_dyn):
        mock_dyn.return_value = 'fooid'
        model_cls = Mock()
        model_cls.pk_field.return_value = 'my_id'
        attr_res.return_value = False
        singular_res.return_value = False
        get_model.return_value = model_cls
        raml_resource = Mock(path='/stories')
        parent_resource = Mock(is_root=False, uid=1)

        res = generators.generate_resource(raml_resource, parent_resource)
        get_model.assert_called_once_with('Story')
        generate_acl.assert_called_once_with(
            model_cls=model_cls, raml_resource=raml_resource)
        mock_dyn.assert_called_once_with(
            raml_resource=raml_resource,
            clean_uri='stories', pk_field='my_id')
        view_attrs.assert_called_once_with(raml_resource, False)
        generate_view.assert_called_once_with(
            model_cls=model_cls,
            attrs=view_attrs(),
            attr_view=False,
            singular=False
        )
        parent_resource.add.assert_called_once_with(
            'story', 'stories',
            id_name='fooid',
            factory=generate_acl(),
            view=generate_view()
        )
        assert res == parent_resource.add()

    @patch('ramses.generators.dynamic_part_name')
    @patch('ramses.generators.singular_subresource')
    @patch('ramses.generators.attr_subresource')
    @patch('ramses.models.get_existing_model')
    @patch('ramses.generators.generate_acl')
    @patch('ramses.generators.resource_view_attrs')
    @patch('ramses.generators.generate_rest_view')
    def test_full_run_singular(
            self, generate_view, view_attrs, generate_acl, get_model,
            attr_res, singular_res, mock_dyn):
        mock_dyn.return_value = 'fooid'
        model_cls = Mock()
        model_cls.pk_field.return_value = 'my_id'
        attr_res.return_value = False
        singular_res.return_value = True
        get_model.return_value = model_cls
        raml_resource = Mock(path='/stories')
        parent_resource = Mock(is_root=False, uid=1)
        parent_resource.view.Model.pk_field.return_value = 'other_id'

        res = generators.generate_resource(raml_resource, parent_resource)
        get_model.assert_called_once_with('Story')
        generate_acl.assert_called_once_with(
            model_cls=parent_resource.view.Model,
            raml_resource=raml_resource)
        assert not mock_dyn.called
        view_attrs.assert_called_once_with(raml_resource, True)
        generate_view.assert_called_once_with(
            model_cls=parent_resource.view.Model,
            attrs=view_attrs(),
            attr_view=False,
            singular=True
        )
        parent_resource.add.assert_called_once_with(
            'story',
            factory=generate_acl(),
            view=generate_view()
        )
        assert res == parent_resource.add()



import pytest
from mock import Mock, patch, call
from pyramid.security import (
    Allow, Deny,
    Everyone, Authenticated,
    ALL_PERMISSIONS)

from ramses import acl


class TestACLHelpers(object):
    methods_map = {'get': 'index', 'post': 'create'}

    def test_methods_to_perms_all_permissions(self):
        perms = acl.methods_to_perms('all,get,post', self.methods_map)
        assert perms is ALL_PERMISSIONS

    def test_methods_to_perms_invalid_perm_name(self):
        with pytest.raises(ValueError) as ex:
            acl.methods_to_perms('foo,post', self.methods_map)
        expected = ("Unknown method name in permissions: "
                    "['foo', 'post']")
        assert expected in str(ex.value)

    def test_methods_to_perms(self):
        perms = acl.methods_to_perms('get', self.methods_map)
        assert perms == ['view']
        perms = acl.methods_to_perms('get,post', self.methods_map)
        assert sorted(perms) == ['create', 'view']

    def test_parse_acl_no_string(self):
        perms = acl.parse_acl('', self.methods_map)
        assert perms == [acl.ALLOW_ALL]

    def test_parse_acl_unknown_action(self):
        with pytest.raises(ValueError) as ex:
            acl.parse_acl('foobar admin all', self.methods_map)
        assert 'Unknown ACL action: foobar' in str(ex.value)

    @patch.object(acl, 'methods_to_perms')
    def test_parse_acl_multiple_values(self, mock_perms):
        mock_perms.return_value = 'Foo'
        perms = acl.parse_acl(
            'allow everyone read,write;allow authenticated all',
            self.methods_map)
        mock_perms.assert_has_calls([
            call(['read', 'write'], self.methods_map),
            call(['all'], self.methods_map),
        ])
        assert sorted(perms) == sorted([
            (Allow, Everyone, 'Foo'),
            (Allow, Authenticated, 'Foo'),
        ])

    @patch.object(acl, 'methods_to_perms')
    def test_parse_acl_special_principal(self, mock_perms):
        mock_perms.return_value = 'Foo'
        perms = acl.parse_acl('allow everyone all', self.methods_map)
        mock_perms.assert_called_once_with(['all'], self.methods_map)
        assert perms == [(Allow, Everyone, 'Foo')]

    @patch.object(acl, 'methods_to_perms')
    def test_parse_acl_group_principal(self, mock_perms):
        mock_perms.return_value = 'Foo'
        perms = acl.parse_acl('allow admin all', self.methods_map)
        mock_perms.assert_called_once_with(['all'], self.methods_map)
        assert perms == [(Allow, 'g:admin', 'Foo')]

    @patch.object(acl, 'resolve_to_callable')
    @patch.object(acl, 'methods_to_perms')
    def test_parse_acl_callable_principal(self, mock_perms, mock_res):
        mock_perms.return_value = 'Foo'
        mock_res.return_value = 'registry callable'
        perms = acl.parse_acl('allow {{my_user}} all', self.methods_map)
        mock_perms.assert_called_once_with(['all'], self.methods_map)
        mock_res.assert_called_once_with('{{my_user}}')
        assert perms == [(Allow, 'registry callable', 'Foo')]


@patch.object(acl, 'parse_acl')
class TestGenerateACL(object):

    def test_no_security(self, mock_parse):
        acl_cls = acl.generate_acl(
            model_cls='Foo',
            raml_resource=Mock(security_schemes=[]),
            es_based=True)
        assert acl_cls.item_model == 'Foo'
        assert issubclass(acl_cls, acl.BaseACL)
        instance = acl_cls(request=None)
        assert instance.es_based
        assert instance._collection_acl == [acl.ALLOW_ALL]
        assert instance._item_acl == [acl.ALLOW_ALL]
        assert not mock_parse.called

    def test_wrong_security_scheme_type(self, mock_parse):
        raml_resource = Mock(security_schemes=[
            Mock(type='x-Foo', settings={'collection': 4, 'item': 7})
        ])
        acl_cls = acl.generate_acl(
            model_cls='Foo',
            raml_resource=raml_resource,
            es_based=False)
        assert not mock_parse.called
        assert acl_cls.item_model == 'Foo'
        assert issubclass(acl_cls, acl.BaseACL)
        instance = acl_cls(request=None)
        assert not instance.es_based
        assert instance._collection_acl == [acl.ALLOW_ALL]
        assert instance._item_acl == [acl.ALLOW_ALL]

    def test_correct_security_scheme(self, mock_parse):
        raml_resource = Mock(security_schemes=[
            Mock(type='x-ACL', settings={'collection': 4, 'item': 7})
        ])
        acl_cls = acl.generate_acl(
            model_cls='Foo',
            raml_resource=raml_resource,
            es_based=False)
        mock_parse.assert_has_calls([
            call(acl_string=4, methods_map=acl.collection_methods),
            call(acl_string=7, methods_map=acl.item_methods),
        ])
        instance = acl_cls(request=None)
        assert instance._collection_acl == mock_parse()
        assert instance._item_acl == mock_parse()
        assert not instance.es_based


class TestBaseACL(object):

    def test_init(self):
        obj = acl.BaseACL(request='Foo')
        assert obj.item_model is None
        assert obj._collection_acl == (acl.ALLOW_ALL,)
        assert obj._item_acl == (acl.ALLOW_ALL,)
        assert obj.request == 'Foo'

    def test_apply_callables_no_callables(self):
        obj = acl.BaseACL('req')
        new_acl = obj._apply_callables(
            acl=[('foo', 'bar', 'baz')],
            methods_map={'zoo': 1},
            obj='obj')
        assert new_acl == [('foo', 'bar', 'baz')]

    @patch.object(acl, 'methods_to_perms')
    def test_apply_callables(self, mock_meth):
        mock_meth.return_value = '123'
        principal = Mock(return_value=(7, 8, 9))
        obj = acl.BaseACL('req')
        new_acl = obj._apply_callables(
            acl=[('foo', principal, 'bar')],
            methods_map={'zoo': 1},
            obj='obj')
        assert new_acl == [(7, 8, '123')]
        principal.assert_called_once_with(
            ace=('foo', principal, 'bar'),
            request='req',
            obj='obj')
        mock_meth.assert_called_once_with(9, {'zoo': 1})

    @patch.object(acl, 'methods_to_perms')
    def test_apply_callables_principal_returns_none(self, mock_meth):
        mock_meth.return_value = '123'
        principal = Mock(return_value=None)
        obj = acl.BaseACL('req')
        new_acl = obj._apply_callables(
            acl=[('foo', principal, 'bar')],
            methods_map={'zoo': 1},
            obj='obj')
        assert new_acl == []
        principal.assert_called_once_with(
            ace=('foo', principal, 'bar'),
            request='req',
            obj='obj')
        assert not mock_meth.called

    @patch.object(acl, 'methods_to_perms')
    def test_apply_callables_principal_returns_list(self, mock_meth):
        mock_meth.return_value = '123'
        principal = Mock(return_value=[(7, 8, 9)])
        obj = acl.BaseACL('req')
        new_acl = obj._apply_callables(
            acl=[('foo', principal, 'bar')],
            methods_map={'zoo': 1},
            obj='obj')
        assert new_acl == [(7, 8, '123')]
        principal.assert_called_once_with(
            ace=('foo', principal, 'bar'),
            request='req',
            obj='obj')
        mock_meth.assert_called_once_with(9, {'zoo': 1})

    def test_apply_callables_functional(self):
        obj = acl.BaseACL('req')
        principal = lambda ace, request, obj: [(Allow, Everyone, 'get')]
        new_acl = obj._apply_callables(
            acl=[(Deny, principal, ALL_PERMISSIONS)],
            methods_map=acl.item_methods,
        )
        assert new_acl == [(Allow, Everyone, ['view'])]

    def test_magic_acl(self):
        obj = acl.BaseACL('req')
        obj._collection_acl = [(1, 2, 3)]
        obj._apply_callables = Mock()
        result = obj.__acl__()
        obj._apply_callables.assert_called_once_with(
            acl=[(1, 2, 3)],
            methods_map=acl.collection_methods
        )
        assert result == obj._apply_callables()

    def test_item_acl(self):
        obj = acl.BaseACL('req')
        obj._item_acl = [(1, 2, 3)]
        obj._apply_callables = Mock()
        result = obj.item_acl('foobar')
        obj._apply_callables.assert_called_once_with(
            acl=[(1, 2, 3)],
            methods_map=acl.item_methods,
            obj='foobar'
        )
        assert result == obj._apply_callables()

    def test_magic_getitem_es_based(self):
        obj = acl.BaseACL('req')
        obj.item_db_id = Mock(return_value=42)
        obj.getitem_es = Mock()
        obj.es_based = True
        obj.__getitem__(1)
        obj.item_db_id.assert_called_once_with(1)
        obj.getitem_es.assert_called_once_with(42)

    def test_magic_getitem_db_based(self):
        obj = acl.BaseACL('req')
        obj.item_db_id = Mock(return_value = 42)
        obj.item_model = Mock()
        obj.item_model.pk_field.return_value = 'id'
        obj.es_based = False
        obj.__getitem__(1)
        obj.item_db_id.assert_called_once_with(1)

    def test_getitem_db(self):
        obj = acl.BaseACL('req')
        obj.item_model = Mock()
        obj.item_model.pk_field.return_value = 'myname'
        obj.item_acl = Mock()
        value = obj['varvar']
        obj.item_model.get.assert_called_once_with(
            __raise=True, myname='varvar')
        obj.item_acl.assert_called_once_with(
            obj.item_model.get())
        assert value.__acl__ == obj.item_acl()
        assert value.__parent__ is obj
        assert value.__name__ == 'varvar'

    @patch('ramses.acl.ES')
    def test_getitem_es(self, mock_es):
        found_obj = Mock()
        es_obj = Mock()
        es_obj.get_resource.return_value = found_obj
        mock_es.return_value = es_obj
        obj = acl.BaseACL('req')
        obj.item_model = Mock(__name__='Foo')
        obj.item_model.pk_field.return_value = 'myname'
        obj.item_acl = Mock()
        value = obj.getitem_es(key='varvar')
        mock_es.assert_called_with('Foo')
        es_obj.get_resource.assert_called_once_with(id='varvar')
        obj.item_acl.assert_called_once_with(found_obj)
        assert value.__acl__ == obj.item_acl()
        assert value.__parent__ is obj
        assert value.__name__ == 'varvar'



import pytest
from mock import Mock, patch

from nefertari.json_httpexceptions import (
    JHTTPNotFound, JHTTPMethodNotAllowed)

from ramses import views


class ViewTestBase(object):
    view_cls = None
    view_kwargs = dict(
        context={},
        _query_params={'foo': 'bar'},
        _json_params={'foo2': 'bar2'},
    )
    request_kwargs = dict(
        method='GET',
        accept=[''],
    )

    def _test_view(self):
        class View(self.view_cls):
            _json_encoder = 'foo'

        request = Mock(**self.request_kwargs)
        return View(request=request, **self.view_kwargs)


class TestBaseView(ViewTestBase):
    view_cls = views.BaseView

    def test_init(self):
        view = self._test_view()
        assert view._query_params['_limit'] == 20

    def test_clean_id_name(self):
        view = self._test_view()
        view._resource = Mock(id_name='foo')
        assert view.clean_id_name == 'foo'
        view._resource = Mock(id_name='foo_bar')
        assert view.clean_id_name == 'bar'

    def test_resolve_kw(self):
        view = self._test_view()
        kwargs = {'foo_bar_qoo': 1, 'arg_val': 4, 'q': 3}
        assert view.resolve_kw(kwargs) == {'bar_qoo': 1, 'val': 4, 'q': 3}

    def test_location(self):
        view = self._test_view()
        view._resource = Mock(id_name='myid', uid='items')
        view._location(Mock(myid=123))
        view.request.route_url.assert_called_once_with(
            'items', myid=123)

    def test_location_split_id(self):
        view = self._test_view()
        view._resource = Mock(id_name='items_myid', uid='items')
        view._location(Mock(myid=123))
        view.request.route_url.assert_called_once_with(
            'items', items_myid=123)

    def test_get_collection_has_parent(self):
        view = self._test_view()
        view._parent_queryset = Mock(return_value=[1, 2, 3])
        view.Model = Mock()
        view.get_collection(name='ok')
        view._parent_queryset.assert_called_once_with()
        view.Model.filter_objects.assert_called_once_with(
            [1, 2, 3], _limit=20, foo='bar', name='ok')

    def test_get_collection_has_parent_empty_queryset(self):
        view = self._test_view()
        view._parent_queryset = Mock(return_value=[])
        view.Model = Mock()
        view.get_collection(name='ok')
        view._parent_queryset.assert_called_once_with()
        view.Model.filter_objects.assert_called_once_with(
            [], _limit=20, foo='bar', name='ok')

    def test_get_collection_no_parent(self):
        view = self._test_view()
        view._parent_queryset = Mock(return_value=None)
        view.Model = Mock()
        view.get_collection(name='ok')
        view._parent_queryset.assert_called_once_with()
        assert not view.Model.filter_objects.called
        view.Model.get_collection.assert_called_once_with(
            _limit=20, foo='bar', name='ok')

    def test_get_item_no_parent(self):
        view = self._test_view()
        view._parent_queryset = Mock(return_value=None)
        view.context = 1
        assert view.get_item(name='wqe') == 1

    def test_get_item_not_found_in_parent(self):
        view = self._test_view()
        view.Model = Mock(__name__='foo')
        view._get_context_key = Mock(return_value='123123')
        view._parent_queryset = Mock(return_value=[2, 3])
        view.context = 1
        with pytest.raises(JHTTPNotFound):
            view.get_item(name='wqe')

    def test_get_item_found_in_parent(self):
        view = self._test_view()
        view._parent_queryset = Mock(return_value=[1, 3])
        view.context = 1
        assert view.get_item(name='wqe') == 1

    def test_get_item_found_in_parent_context_callable(self):
        func = lambda x: x
        view = self._test_view()
        view._parent_queryset = Mock(return_value=[func, 3])
        view.reload_context = Mock()
        view.context = func
        assert view.get_item(name='wqe') is view.context
        view.reload_context.assert_called_once_with(
            es_based=False, name='wqe')

    def test_get_context_key(self):
        view = self._test_view()
        view._resource = Mock(id_name='foo')
        assert view._get_context_key(foo='bar') == 'bar'

    def test_parent_queryset(self):
        from pyramid.config import Configurator
        from ramses.acl import BaseACL
        config = Configurator()
        config.include('nefertari')
        root = config.get_root_resource()
        user = root.add(
            'user', 'users', id_name='username',
            view=views.BaseView, factory=BaseACL)
        user.add(
            'story', 'stories', id_name='prof_id',
            view=views.BaseView, factory=BaseACL)
        view_cls = root.resource_map['user:story'].view
        view_cls._json_encoder = 'foo'

        request = Mock(
            registry=Mock(),
            path='/foo/foo',
            matchdict={'username': 'user12', 'prof_id': 4},
            accept=[''], method='GET'
        )
        request.params.mixed.return_value = {'foo1': 'bar1'}
        request.blank.return_value = request
        stories_view = view_cls(
            request=request,
            context={},
            _query_params={'foo1': 'bar1'},
            _json_params={'foo2': 'bar2'},)

        parent_view = stories_view._resource.parent.view
        with patch.object(parent_view, 'get_item') as get_item:
            parent_view.get_item = get_item
            result = stories_view._parent_queryset()
            get_item.assert_called_once_with(username='user12')
            assert result == get_item().stories

    def test_reload_context(self):
        class Factory(dict):
            item_model = None

            def __getitem__(self, key):
                return key

        view = self._test_view()
        view._factory = Factory
        view._get_context_key = Mock(return_value='foo')
        view.reload_context(es_based=False, arg='asd')
        view._get_context_key.assert_called_once_with(arg='asd')
        assert view.context == 'foo'


class TestCollectionView(ViewTestBase):
    view_cls = views.CollectionView

    def test_index(self):
        view = self._test_view()
        view.get_collection = Mock()
        resp = view.index(foo='bar')
        view.get_collection.assert_called_once_with()
        assert resp == view.get_collection()

    def test_show(self):
        view = self._test_view()
        view.get_item = Mock()
        resp = view.show(foo='bar')
        view.get_item.assert_called_once_with(foo='bar')
        assert resp == view.get_item()

    def test_create(self):
        view = self._test_view()
        view.request.registry._root_resources = {
            'foo': Mock(auth=False)
        }
        view.Model = Mock()
        obj = Mock()
        obj.to_dict.return_value = {'id': 1}
        view.Model().save.return_value = obj
        view._location = Mock(return_value='/sadasd')
        resp = view.create(foo='bar')
        view.Model.assert_called_with(foo2='bar2')
        view.Model().save.assert_called_with(view.request)
        assert resp == view.Model().save()

    def test_update(self):
        view = self._test_view()
        view.get_item = Mock()
        view._location = Mock(return_value='/sadasd')
        resp = view.update(foo=1)
        view.get_item.assert_called_once_with(foo=1)
        view.get_item().update.assert_called_once_with(
            {'foo2': 'bar2'}, view.request)
        assert resp == view.get_item().update()

    def test_replace(self):
        view = self._test_view()
        view.update = Mock()
        resp = view.replace(foo=1)
        view.update.assert_called_once_with(foo=1)
        assert resp == view.update()

    def test_delete(self):
        view = self._test_view()
        view.get_item = Mock()
        resp = view.delete(foo=1)
        view.get_item.assert_called_once_with(foo=1)
        view.get_item().delete.assert_called_once_with(
            view.request)
        assert resp is None

    def test_delete_many_needs_confirm(self):
        view = self._test_view()
        view.Model = Mock()
        view.get_collection = Mock()
        view.needs_confirmation = Mock(return_value=True)
        resp = view.delete_many(foo=1)
        view.get_collection.assert_called_once_with()
        assert resp == view.get_collection()

    def test_delete_many(self):
        view = self._test_view()
        view.Model = Mock(__name__='Mock')
        view.Model._delete_many.return_value = 123
        view.get_collection = Mock()
        view.needs_confirmation = Mock(return_value=False)
        resp = view.delete_many(foo=1)
        view.get_collection.assert_called_once_with()
        view.Model._delete_many.assert_called_once_with(
            view.get_collection(), view.request)
        assert resp == 123

    def test_update_many(self):
        view = self._test_view()
        view.Model = Mock(__name__='Mock')
        view.Model._update_many.return_value = 123
        view.get_collection = Mock()
        resp = view.update_many(qoo=1)
        view.get_collection.assert_called_once_with(_limit=20, foo='bar')
        view.Model._update_many.assert_called_once_with(
            view.get_collection(), {'foo2': 'bar2'},
            view.request)
        assert resp == 123


class TestESBaseView(ViewTestBase):
    view_cls = views.ESBaseView

    def test_parent_queryset_es(self):
        from pyramid.config import Configurator
        from ramses.acl import BaseACL
        config = Configurator()
        config.include('nefertari')
        root = config.get_root_resource()
        user = root.add(
            'user', 'users', id_name='username',
            view=views.ESBaseView, factory=BaseACL)
        user.add(
            'story', 'stories', id_name='prof_id',
            view=views.ESBaseView, factory=BaseACL)
        view_cls = root.resource_map['user:story'].view
        view_cls._json_encoder = 'foo'

        request = Mock(
            registry=Mock(),
            path='/foo/foo',
            matchdict={'username': 'user12', 'prof_id': 4},
            accept=[''], method='GET'
        )
        request.params.mixed.return_value = {'foo1': 'bar1'}
        request.blank.return_value = request
        stories_view = view_cls(
            request=request,
            context={},
            _query_params={'foo1': 'bar1'},
            _json_params={'foo2': 'bar2'},)

        parent_view = stories_view._resource.parent.view
        with patch.object(parent_view, 'get_item_es') as get_item_es:
            parent_view.get_item_es = get_item_es
            result = stories_view._parent_queryset_es()
            get_item_es.assert_called_once_with(username='user12')
            assert result == get_item_es().stories

    def test_get_es_object_ids(self):
        view = self._test_view()
        view._resource = Mock(id_name='foobar')
        objects = [Mock(foobar=4), Mock(foobar=7)]
        assert sorted(view.get_es_object_ids(objects)) == ['4', '7']

    @patch('nefertari.elasticsearch.ES')
    def test_get_collection_es_no_parent(self, mock_es):
        view = self._test_view()
        view._parent_queryset_es = Mock(return_value=None)
        view.Model = Mock(__name__='Foo')
        view.get_collection_es(arg=1)
        mock_es.assert_called_once_with('Foo')
        mock_es().get_collection.assert_called_once_with(
            _limit=20, foo='bar')

    @patch('nefertari.elasticsearch.ES')
    def test_get_collection_es_parent_no_obj_ids(self, mock_es):
        view = self._test_view()
        view._parent_queryset_es = Mock(return_value=[1, 2])
        view.Model = Mock(__name__='Foo')
        view.get_es_object_ids = Mock(return_value=None)
        result = view.get_collection_es(arg=1)
        assert not mock_es().get_collection.called
        assert result == []

    @patch('nefertari.elasticsearch.ES')
    def test_get_collection_es_parent_with_ids(self, mock_es):
        view = self._test_view()
        view._parent_queryset_es = Mock(return_value=['obj1', 'obj2'])
        view.Model = Mock(__name__='Foo')
        view.get_es_object_ids = Mock(return_value=[1, 2])
        view.get_collection_es(arg=7)
        view.get_es_object_ids.assert_called_once_with(['obj1', 'obj2'])
        mock_es().get_collection.assert_called_once_with(
            _limit=20, foo='bar', id=[1, 2])

    def test_get_item_es_no_parent(self):
        view = self._test_view()
        view._get_context_key = Mock(return_value=1)
        view._parent_queryset_es = Mock(return_value=None)
        view.reload_context = Mock()
        view.context = 'foo'
        resp = view.get_item_es(a=4)
        view._get_context_key.assert_called_once_with(a=4)
        view._parent_queryset_es.assert_called_once_with()
        assert not view.reload_context.called
        assert resp == 'foo'

    def test_get_item_es_matching_id(self):
        view = self._test_view()
        view._get_context_key = Mock(return_value=1)
        view._parent_queryset_es = Mock(return_value=['obj1', 'obj2'])
        view.get_es_object_ids = Mock(return_value=[1, 2])
        view.reload_context = Mock()
        view.context = 'foo'
        resp = view.get_item_es(a=4)
        view.get_es_object_ids.assert_called_once_with(['obj1', 'obj2'])
        view._get_context_key.assert_called_once_with(a=4)
        view._parent_queryset_es.assert_called_once_with()
        assert not view.reload_context.called
        assert resp == 'foo'

    def test_get_item_es_not_matching_id(self):
        view = self._test_view()
        view._get_context_key = Mock(return_value=1)
        view._parent_queryset_es = Mock(return_value=['obj1', 'obj2'])
        view.get_es_object_ids = Mock(return_value=[2, 3])
        view.reload_context = Mock()
        view.Model = Mock(__name__='Foo')
        view.context = 'foo'
        with pytest.raises(JHTTPNotFound) as ex:
            view.get_item_es(a=4)
        assert 'Foo(id=1) resource not found' in str(ex.value)

    def test_get_item_es_callable_context(self):
        view = self._test_view()
        view._get_context_key = Mock(return_value=1)
        view._parent_queryset_es = Mock(return_value=['obj1', 'obj2'])
        view.get_es_object_ids = Mock(return_value=[1, 2])
        view.reload_context = Mock()
        view.context = lambda x: x
        resp = view.get_item_es(a=4)
        view.reload_context.assert_called_once_with(es_based=True, a=4)
        assert resp == view.context


class TestESCollectionView(ViewTestBase):
    view_cls = views.ESCollectionView

    def test_index(self):
        view = self._test_view()
        view.aggregate = Mock(side_effect=KeyError)
        view.get_collection_es = Mock()
        resp = view.index(foo=1)
        view.get_collection_es.assert_called_once_with(foo=1)
        assert resp == view.get_collection_es()

    def test_show(self):
        view = self._test_view()
        view.get_item_es = Mock()
        resp = view.show(foo=1)
        view.get_item_es.assert_called_once_with(foo=1)
        assert resp == view.get_item_es()

    def test_update(self):
        view = self._test_view()
        view.get_item = Mock()
        view.reload_context = Mock()
        view._location = Mock(return_value='/sadasd')
        resp = view.update(foo=1)
        view.reload_context.assert_called_once_with(es_based=False, foo=1)
        view.get_item.assert_called_once_with(foo=1)
        view.get_item().update.assert_called_once_with(
            {'foo2': 'bar2'}, view.request)
        assert resp == view.get_item().update()

    def test_replace(self):
        view = self._test_view()
        view.update = Mock()
        resp = view.replace(foo=1)
        view.update.assert_called_once_with(foo=1)
        assert resp == view.update()

    def test_get_dbcollection_with_es(self):
        view = self._test_view()
        view._query_params['_limit'] = 50
        view.get_collection_es = Mock(return_value=[1, 2])
        view.Model = Mock()
        result = view.get_dbcollection_with_es(foo='bar')
        view.get_collection_es.assert_called_once_with(foo='bar')
        view.Model.filter_objects.assert_called_once_with([1, 2])
        assert result == view.Model.filter_objects()

    def test_delete_many_need_confirmation(self):
        view = self._test_view()
        view.needs_confirmation = Mock(return_value=True)
        view.Model = Mock()
        view.get_dbcollection_with_es = Mock()
        result = view.delete_many(foo=1)
        view.get_dbcollection_with_es.assert_called_once_with(foo=1)
        assert result == view.get_dbcollection_with_es()
        assert not view.Model._delete_many.called

    def test_delete_many(self):
        view = self._test_view()
        view.needs_confirmation = Mock(return_value=False)
        view.Model = Mock(__name__='Foo')
        view.Model._delete_many.return_value = 123
        view.get_dbcollection_with_es = Mock()
        result = view.delete_many(foo=1)
        view.get_dbcollection_with_es.assert_called_once_with(foo=1)
        view.Model._delete_many.assert_called_once_with(
            view.get_dbcollection_with_es(), view.request)
        assert result == 123

    def test_update_many(self):
        view = self._test_view()
        view.needs_confirmation = Mock(return_value=False)
        view.Model = Mock(__name__='Foo')
        view.Model._update_many.return_value = 123
        view.get_dbcollection_with_es = Mock()
        result = view.update_many(foo=1)
        view.get_dbcollection_with_es.assert_called_once_with(foo=1)
        view.Model._update_many.assert_called_once_with(
            view.get_dbcollection_with_es(), {'foo2': 'bar2'},
            view.request)
        assert result == 123


class TestItemSubresourceBaseView(ViewTestBase):
    view_cls = views.ItemSubresourceBaseView

    def test_get_context_key(self):
        view = self._test_view()
        parent = Mock(id_name='foobar')
        resource = Mock()
        resource.parent = parent
        view._resource = resource
        assert view._get_context_key(foobar=1, foo=2) == '1'

    def test_get_item(self):
        view = self._test_view()
        view._parent_queryset = Mock(return_value=[1, 2])
        view.reload_context = Mock()
        view.context = 1
        assert view.get_item(foo=4) == 1
        view._parent_queryset.assert_called_once_with()
        view.reload_context.assert_called_once_with(es_based=False, foo=4)


class TestItemAttributeView(ViewTestBase):
    view_cls = views.ItemAttributeView
    request_kwargs = dict(
        method='GET',
        accept=[''],
        path='user/1/settings'
    )

    def test_init(self):
        view = self._test_view()
        assert view.value_type is None
        assert view.unique
        assert view.attr == 'settings'

    def test_index(self):
        view = self._test_view()
        view.get_item = Mock()
        resp = view.index(foo=1)
        view.get_item.assert_called_once_with(foo=1)
        assert resp == view.get_item().settings

    def test_create(self):
        view = self._test_view()
        view.get_item = Mock()
        resp = view.create(foo=1)
        view.get_item.assert_called_once_with(foo=1)
        obj = view.get_item()
        obj.update_iterables.assert_called_once_with(
            {'foo2': 'bar2'}, 'settings',
            unique=True, value_type=None,
            request=view.request)
        assert resp == obj.settings


class TestItemSingularView(ViewTestBase):
    view_cls = views.ItemSingularView
    request_kwargs = dict(
        method='GET',
        accept=[''],
        path='user/1/profile',
        url='http://example.com',
    )

    def test_init(self):
        view = self._test_view()
        assert view.attr == 'profile'

    def test_show(self):
        view = self._test_view()
        view.get_item = Mock()
        resp = view.show(foo=1)
        view.get_item.assert_called_once_with(foo=1)
        assert resp == view.get_item().profile

    def test_create(self):
        view = self._test_view()
        view.request.registry._root_resources = {
            'foo': Mock(auth=False)
        }
        view.get_item = Mock()
        view._singular_model = Mock()
        resp = view.create(foo=1)
        view.get_item.assert_called_once_with(foo=1)
        view._singular_model.assert_called_once_with(foo2='bar2')
        child = view._singular_model()
        child.save.assert_called_once_with(view.request)
        parent = view.get_item()
        parent.update.assert_called_once_with(
            {'profile': child.save()}, view.request)
        assert resp == child.save()

    def test_update(self):
        view = self._test_view()
        view.get_item = Mock()
        resp = view.update(foo=1)
        view.get_item.assert_called_once_with(foo=1)
        child = view.get_item().profile
        child.update.assert_called_once_with(
            {'foo2': 'bar2'}, view.request)
        assert resp == child

    def test_replace(self):
        view = self._test_view()
        view.update = Mock()
        resp = view.replace(foo=1)
        view.update.assert_called_once_with(foo=1)
        assert resp == view.update()

    def test_delete(self):
        view = self._test_view()
        view.attr = 'profile'
        view.get_item = Mock()
        resp = view.delete(foo=1)
        assert resp is None
        view.get_item.assert_called_once_with(foo=1)
        parent = view.get_item()
        parent.profile.delete.assert_called_once_with(
            view.request)


class TestRestViewGeneration(object):

    @patch('ramses.views.ESCollectionView._run_init_actions')
    def test_only_provided_attrs_are_available(self, run_init):
        view_cls = views.generate_rest_view(
            model_cls='foo', attrs=['show', 'foobar'],
            es_based=True, attr_view=False, singular=False)
        view_cls._json_encoder = 'foo'
        assert issubclass(view_cls, views.ESCollectionView)
        request = Mock(**ViewTestBase.request_kwargs)
        view = view_cls(request=request, **ViewTestBase.view_kwargs)
        assert not hasattr(view_cls, 'foobar')

        try:
            view.show()
        except JHTTPMethodNotAllowed:
            raise Exception('Unexpected error')
        except Exception:
            pass
        with pytest.raises(JHTTPMethodNotAllowed):
            view.delete_many()
        with pytest.raises(JHTTPMethodNotAllowed):
            view.create()
        with pytest.raises(JHTTPMethodNotAllowed):
            view.delete()
        with pytest.raises(JHTTPMethodNotAllowed):
            view.update_many()
        with pytest.raises(JHTTPMethodNotAllowed):
            view.index()

    def test_singular_view(self):
        view_cls = views.generate_rest_view(
            model_cls='foo', attrs=['show'],
            es_based=True, attr_view=False, singular=True)
        view_cls._json_encoder = 'foo'
        assert issubclass(view_cls, views.ItemSingularView)

    def test_attribute_view(self):
        view_cls = views.generate_rest_view(
            model_cls='foo', attrs=['show'],
            es_based=True, attr_view=True, singular=False)
        view_cls._json_encoder = 'foo'
        assert issubclass(view_cls, views.ItemAttributeView)

    def test_escollection_view(self):
        view_cls = views.generate_rest_view(
            model_cls='foo', attrs=['show'],
            es_based=True, attr_view=False, singular=False)
        view_cls._json_encoder = 'foo'
        assert issubclass(view_cls, views.ESCollectionView)
        assert issubclass(view_cls, views.CollectionView)

    def test_dbcollection_view(self):
        view_cls = views.generate_rest_view(
            model_cls='foo', attrs=['show'],
            es_based=False, attr_view=False, singular=False)
        view_cls._json_encoder = 'foo'
        assert not issubclass(view_cls, views.ESCollectionView)
        assert issubclass(view_cls, views.CollectionView)

    def test_default_values(self):
        view_cls = views.generate_rest_view(
            model_cls='foo', attrs=['show'])
        view_cls._json_encoder = 'foo'
        assert issubclass(view_cls, views.ESCollectionView)
        assert issubclass(view_cls, views.CollectionView)
        assert view_cls.Model == 'foo'



import pytest


@pytest.fixture
def clear_registry(request):
    from ramses import registry
    registry.registry.clear()


@pytest.fixture
def engine_mock(request):
    import nefertari
    from mock import Mock

    class BaseDocument(object):
        pass

    class ESBaseDocument(object):
        pass

    original_engine = nefertari.engine
    nefertari.engine = Mock()
    nefertari.engine.BaseDocument = BaseDocument
    nefertari.engine.ESBaseDocument = ESBaseDocument

    def clear():
        nefertari.engine = original_engine
    request.addfinalizer(clear)

    return nefertari.engine



import logging

from inflection import singularize

from .views import generate_rest_view
from .acl import generate_acl
from .utils import (
    is_dynamic_uri, resource_view_attrs, generate_model_name,
    dynamic_part_name, attr_subresource, singular_subresource,
    get_static_parent)


log = logging.getLogger(__name__)


def _get_nefertari_parent_resource(
        raml_resource, generated_resources, default):
    parent_raml_res = get_static_parent(raml_resource)
    if parent_raml_res is not None:
        parent_resource = generated_resources.get(parent_raml_res.path)
        return parent_resource or default
    return default


def generate_resource(raml_resource, parent_resource):
    """ Perform complete one resource configuration process

    This function generates: ACL, view, route, resource, database
    model for a given `raml_resource`. New nefertari resource is
    attached to `parent_resource` class which is an instance of
    `nefertari.resource.Resource`.

    Things to consider:
      * Top-level resources must be collection names.
      * No resources are explicitly created for dynamic (ending with '}')
        RAML resources as they are implicitly processed by parent collection
        resources.
      * Resource nesting must look like collection/id/collection/id/...
      * Only part of resource path after last '/' is taken into account,
        thus each level of resource nesting should add one more path
        element. E.g. /stories -> /stories/{id} and not
        /stories -> /stories/mystories/{id}. Latter route will be generated
        at /stories/{id}.

    :param raml_resource: Instance of Instance of
        ramlfications.raml.ResourceNode.
    :param parent_resource: Parent nefertari resource object.
    """
    from .models import get_existing_model

    # Don't generate resources for dynamic routes as they are already
    # generated by their parent
    resource_uri = raml_resource.path.split('/')[-1].strip()
    if is_dynamic_uri(resource_uri):
        if parent_resource.is_root:
            raise Exception("Top-level resources can't be dynamic and must "
                            "represent collections instead")
        return

    clean_uri = route_name = resource_uri.strip('/')
    log.info('Configuring resource: `{}`. Parent: `{}`'.format(
        route_name, parent_resource.uid or 'root'))

    # Get DB model. If this is an attribute or singular resource,
    # we don't need to get model
    is_singular = singular_subresource(raml_resource, route_name)
    is_attr_res = attr_subresource(raml_resource, route_name)
    if not parent_resource.is_root and (is_attr_res or is_singular):
        model_cls = parent_resource.view.Model
    else:
        model_name = generate_model_name(route_name)
        model_cls = get_existing_model(model_name)

    resource_kwargs = {}

    # Generate ACL
    log.info('Generating ACL for `{}`'.format(route_name))
    resource_kwargs['factory'] = generate_acl(
        model_cls=model_cls,
        raml_resource=raml_resource)

    # Generate dynamic part name
    if not is_singular:
        resource_kwargs['id_name'] = dynamic_part_name(
            raml_resource=raml_resource,
            clean_uri=clean_uri,
            pk_field=model_cls.pk_field())

    # Generate REST view
    log.info('Generating view for `{}`'.format(route_name))
    view_attrs = resource_view_attrs(raml_resource, is_singular)
    resource_kwargs['view'] = generate_rest_view(
        model_cls=model_cls,
        attrs=view_attrs,
        attr_view=is_attr_res,
        singular=is_singular,
    )

    # In case of singular resource, model still needs to be generated,
    # but we store it on a different view attribute
    if is_singular:
        model_name = generate_model_name(route_name)
        resource_kwargs['view']._singular_model = get_existing_model(
            model_name)

    # Create new nefertari resource
    log.info('Creating new resource for `{}`'.format(route_name))
    resource_args = (singularize(clean_uri),)

    if not is_singular:
        resource_args += (clean_uri,)

    return parent_resource.add(*resource_args, **resource_kwargs)


def generate_server(raml_root, config):
    """ Handle server generation process.

    :param raml_root: Instance of ramlfications.raml.RootNode.
    :param config: Pyramid Configurator instance.
    """
    log.info('Server generation started')

    if not raml_root.resources:
        return

    root_resource = config.get_root_resource()
    generated_resources = {}

    for raml_resource in raml_root.resources:
        if raml_resource.path in generated_resources:
            continue

        # Get Nefertari parent resource
        parent_resource = _get_nefertari_parent_resource(
            raml_resource, generated_resources, root_resource)

        # Get generated resource and store it
        new_resource = generate_resource(raml_resource, parent_resource)
        if new_resource is not None:
            generated_resources[raml_resource.path] = new_resource


def generate_models(config, raml_resources):
    """ Generate model for each resource in :raml_resources:

    The DB model name is generated using singular titled version of current
    resource's url. E.g. for resource under url '/stories', model with
    name 'Story' will be generated.

    :param config: Pyramid Configurator instance.
    :param raml_resources: List of ramlfications.raml.ResourceNode.
    """
    from .models import handle_model_generation
    if not raml_resources:
        return

    for raml_resource in raml_resources:
        # No need to generate models for dynamic resource
        if is_dynamic_uri(raml_resource.path):
            continue

        # Since POST resource must define schema use only POST
        # resources to generate models
        if raml_resource.method.upper() != 'POST':
            continue

        # Generate DB model
        # If this is an attribute resource we don't need to generate model
        route_name = raml_resource.path.split('/')[-1]
        route_name = route_name.strip('/')
        if not attr_subresource(raml_resource, route_name):
            log.info('Configuring model for route `{}`'.format(route_name))
            model_cls, is_auth_model = handle_model_generation(
                raml_resource, route_name)
            if is_auth_model:
                config.registry.auth_model = model_cls



import logging

from nefertari import engine

from .utils import (
    resolve_to_callable, is_callable_tag,
    resource_schema, generate_model_name)
from . import registry


log = logging.getLogger(__name__)

"""
Map of RAML types names to nefertari.engine fields.

"""
type_fields = {
    'string':           engine.StringField,
    'float':            engine.FloatField,
    'integer':          engine.IntegerField,
    'boolean':          engine.BooleanField,
    'datetime':         engine.DateTimeField,
    'file':             engine.BinaryField,
    'relationship':     engine.Relationship,
    'dict':             engine.DictField,
    'foreign_key':      engine.ForeignKeyField,
    'big_integer':      engine.BigIntegerField,
    'date':             engine.DateField,
    'choice':           engine.ChoiceField,
    'interval':         engine.IntervalField,
    'decimal':          engine.DecimalField,
    'pickle':           engine.PickleField,
    'small_integer':    engine.SmallIntegerField,
    'text':             engine.TextField,
    'time':             engine.TimeField,
    'unicode':          engine.UnicodeField,
    'unicode_text':     engine.UnicodeTextField,
    'id_field':         engine.IdField,
    'list':             engine.ListField,
}


def get_existing_model(model_name):
    """ Try to find existing model class named `model_name`.

    :param model_name: String name of the model class.
    """
    try:
        model_cls = engine.get_document_cls(model_name)
        log.debug('Model `{}` already exists. Using existing one'.format(
            model_name))
        return model_cls
    except ValueError:
        log.debug('Model `{}` does not exist'.format(model_name))


def prepare_relationship(field_name, model_name, raml_resource):
    """ Create referenced model if it doesn't exist.

    When preparing a relationship, we check to see if the model that will be
    referenced already exists. If not, it is created so that it will be possible
    to use it in a relationship. Thus the first usage of this model in RAML file
    must provide its schema in POST method resource body schema.

    :param field_name: Name of the field that should become a `Relationship`.
    :param model_name: Name of model which should be generated.
    :param raml_resource: Instance of ramlfications.raml.ResourceNode for
        which :model_name: will be defined.
    """
    if get_existing_model(model_name) is None:
        for res in raml_resource.root.resources:
            if res.method.upper() != 'POST':
                continue
            if res.path.endswith('/' + field_name):
                break
        else:
            raise ValueError('Model `{}` used in relationship `{}` is not '
                             'defined'.format(model_name, field_name))
        setup_data_model(res, model_name)


def generate_model_cls(schema, model_name, raml_resource, es_based=True):
    """ Generate model class.

    Engine DB field types are determined using `type_fields` and only those
    types may be used.

    :param schema: Model schema dict parsed from RAML.
    :param model_name: String that is used as new model's name.
    :param raml_resource: Instance of ramlfications.raml.ResourceNode.
    :param es_based: Boolean indicating if generated model should be a
        subclass of Elasticsearch-based document class or not.
        It True, ESBaseDocument is used; BaseDocument is used otherwise.
        Defaults to True.
    """
    from nefertari.authentication.models import AuthModelMethodsMixin
    base_cls = engine.ESBaseDocument if es_based else engine.BaseDocument
    model_name = str(model_name)
    metaclass = type(base_cls)
    auth_model = schema.get('_auth_model', False)
    if auth_model:
        bases = (AuthModelMethodsMixin, base_cls)
    else:
        bases = (base_cls,)
    attrs = {
        '__tablename__': model_name.lower(),
        '_public_fields': schema.get('_public_fields') or [],
        '_auth_fields': schema.get('_auth_fields') or [],
        '_nested_relationships': schema.get('_nested_relationships') or [],
    }
    # Generate fields from properties
    properties = schema.get('properties', {})
    for field_name, props in properties.items():
        if field_name in attrs:
            continue

        db_settings = props.get('_db_settings')
        if db_settings is None:
            continue

        field_kwargs = db_settings.copy()
        field_kwargs['required'] = bool(field_kwargs.get('required'))

        for default_attr_key in ('default', 'onupdate'):
            value = field_kwargs.get(default_attr_key)
            if is_callable_tag(value):
                field_kwargs[default_attr_key] = resolve_to_callable(value)

        for processor_key in ('before_validation',
                              'after_validation',
                              'backref_before_validation',
                              'backref_after_validation'):
            if processor_key in field_kwargs:
                processors = field_kwargs.get(processor_key, [])
                field_kwargs[processor_key] = [
                    resolve_to_callable(name) for name in processors]

        type_name = (
            field_kwargs.pop('type', 'string') or 'string').lower()
        if type_name not in type_fields:
            raise ValueError('Unknown type: {}'.format(type_name))

        field_cls = type_fields[type_name]

        if field_cls is engine.Relationship:
            prepare_relationship(
                field_name, field_kwargs['document'], raml_resource)
        if field_cls is engine.ForeignKeyField:
            key = 'ref_column_type'
            field_kwargs[key] = type_fields[field_kwargs[key]]
        if field_cls is engine.ListField:
            key = 'item_type'
            field_kwargs[key] = type_fields[field_kwargs[key]]

        attrs[field_name] = field_cls(**field_kwargs)

    # Update model definition with methods and variables defined in registry
    attrs.update(registry.mget(model_name))

    # Generate new model class
    return metaclass(model_name, bases, attrs), auth_model


def setup_data_model(raml_resource, model_name):
    """ Setup storage/data model and return generated model class.

    Process follows these steps:
      * Resource schema is found and restructured by `resource_schema`.
      * Model class is generated from properties dict using util function
        `generate_model_cls`.

    :param raml_resource: Instance of ramlfications.raml.ResourceNode.
    :param model_name: String representing model name.
    """
    from .models import generate_model_cls, get_existing_model
    model_cls = get_existing_model(model_name)
    if model_cls is not None:
        return model_cls, False

    schema = resource_schema(raml_resource)
    if not schema:
        raise Exception('Missing schema for model `{}`'.format(model_name))

    log.info('Generating model class `{}`'.format(model_name))
    return generate_model_cls(
        schema=schema,
        model_name=model_name,
        raml_resource=raml_resource,
    )


def handle_model_generation(raml_resource, route_name):
    """ Generates model name and runs `setup_data_model` to get
    or generate actual model class.

    :param raml_resource: Instance of ramlfications.raml.ResourceNode.
    :param route_name: String name of the resource.
    """
    model_name = generate_model_name(route_name)
    try:
        return setup_data_model(raml_resource, model_name)
    except ValueError as ex:
        raise ValueError('{}: {}'.format(model_name, str(ex)))



import pytest
from mock import Mock, patch

from ramses import utils


class TestUtils(object):

    def test_contenttypes(self):
        assert utils.ContentTypes.JSON == 'application/json'
        assert utils.ContentTypes.TEXT_XML == 'text/xml'
        assert utils.ContentTypes.MULTIPART_FORMDATA == \
            'multipart/form-data'
        assert utils.ContentTypes.FORM_URLENCODED == \
            'application/x-www-form-urlencoded'

    def test_convert_schema_json(self):
        schema = utils.convert_schema({'foo': 'bar'}, 'application/json')
        assert schema == {'foo': 'bar'}

    def test_convert_schema_json_error(self):
        with pytest.raises(TypeError) as ex:
            utils.convert_schema('foo', 'application/json')
        assert 'Schema is not a valid JSON' in str(ex.value)

    def test_convert_schema_xml(self):
        assert utils.convert_schema({'foo': 'bar'}, 'text/xml') is None

    def test_is_dynamic_uri(self):
        assert utils.is_dynamic_uri('/{id}')
        assert not utils.is_dynamic_uri('/collection')

    def test_clean_dynamic_uri(self):
        clean = utils.clean_dynamic_uri('/{item_id}')
        assert clean == 'item_id'

    def test_generate_model_name(self):
        model_name = utils.generate_model_name('/collectionitems')
        assert model_name == 'Collectionitem'

    @patch.object(utils, 'get_resource_children')
    def test_dynamic_part_name(self, get_children):
        get_children.return_value = [
            Mock(path='/items'), Mock(path='/{myid}')]
        resource = Mock()
        part_name = utils.dynamic_part_name(
            resource, 'stories', 'default_id')
        assert part_name == 'stories_myid'
        get_children.assert_called_once_with(resource)

    @patch.object(utils, 'get_resource_children')
    def test_dynamic_part_name_no_dynamic(self, get_children):
        get_children.return_value = [Mock(path='/items')]
        resource = Mock()
        part_name = utils.dynamic_part_name(
            resource, 'stories', 'default_id')
        assert part_name == 'stories_default_id'
        get_children.assert_called_once_with(resource)

    @patch.object(utils, 'get_resource_children')
    def test_dynamic_part_name_no_resources(self, get_children):
        get_children.return_value = []
        resource = Mock(resources=None)
        part_name = utils.dynamic_part_name(
            resource, 'stories', 'default_id')
        assert part_name == 'stories_default_id'
        get_children.assert_called_once_with(resource)

    def test_extract_dynamic_part(self):
        assert utils.extract_dynamic_part('/stories/{id}/foo') == 'id'
        assert utils.extract_dynamic_part('/stories/{id}') == 'id'

    def test_extract_dynamic_part_fail(self):
        assert utils.extract_dynamic_part('/stories/id') is None

    def _get_mock_method_resources(self, *methods):
        return [Mock(method=meth) for meth in methods]

    @patch.object(utils, 'get_resource_children')
    @patch.object(utils, 'get_resource_siblings')
    def test_resource_view_attrs_no_dynamic_subres(self, get_sib, get_child):
        get_child.return_value = []
        get_sib.return_value = self._get_mock_method_resources(
            'get', 'post', 'put', 'patch', 'delete')
        resource = Mock()
        attrs = utils.resource_view_attrs(resource, singular=False)
        get_sib.assert_called_once_with(resource)
        get_child.assert_called_once_with(resource)
        assert attrs == set(['create', 'delete_many', 'index', 'update_many'])

    @patch.object(utils, 'get_resource_children')
    @patch.object(utils, 'get_resource_siblings')
    def test_resource_view_attrs_dynamic_subres(self, get_sib, get_child):
        get_child.return_value = self._get_mock_method_resources(
            'get', 'put', 'patch', 'delete')
        get_sib.return_value = self._get_mock_method_resources(
            'get', 'post', 'put', 'patch', 'delete')
        resource = Mock()
        attrs = utils.resource_view_attrs(resource, singular=False)
        get_sib.assert_called_once_with(resource)
        get_child.assert_called_once_with(resource)
        assert attrs == set([
            'create', 'delete_many', 'index', 'update_many',
            'show', 'update', 'delete', 'replace'
        ])

    @patch.object(utils, 'get_resource_children')
    @patch.object(utils, 'get_resource_siblings')
    def test_resource_view_attrs_singular(self, get_sib, get_child):
        get_child.return_value = []
        get_sib.return_value = self._get_mock_method_resources(
            'get', 'post', 'put', 'patch', 'delete')
        resource = Mock()
        attrs = utils.resource_view_attrs(resource, singular=True)
        get_sib.assert_called_once_with(resource)
        get_child.assert_called_once_with(resource)
        assert attrs == set(['create', 'delete', 'show', 'update', 'replace'])

    @patch.object(utils, 'get_resource_children')
    @patch.object(utils, 'get_resource_siblings')
    def test_resource_view_attrs_no_subresources(self, get_sib, get_child):
        child_res = self._get_mock_method_resources('get')
        child_res[0].path = '/items'
        get_child.return_value = child_res
        get_sib.return_value = self._get_mock_method_resources(
            'get', 'post', 'put', 'patch', 'delete')
        resource = Mock()
        attrs = utils.resource_view_attrs(resource, singular=False)
        get_sib.assert_called_once_with(resource)
        get_child.assert_called_once_with(resource)
        assert attrs == set(['create', 'delete_many', 'index', 'update_many'])

    @patch.object(utils, 'get_resource_children')
    @patch.object(utils, 'get_resource_siblings')
    def test_resource_view_attrs_no_methods(self, get_sib, get_child):
        get_sib.return_value = []
        get_child.return_value = []
        resource = Mock()
        attrs = utils.resource_view_attrs(resource, singular=False)
        get_sib.assert_called_once_with(resource)
        get_child.assert_called_once_with(resource)
        assert attrs == set()

    @patch.object(utils, 'get_resource_children')
    @patch.object(utils, 'get_resource_siblings')
    def test_resource_view_attrs_not_supported_method(
            self, get_sib, get_child):
        get_sib.return_value = []
        get_child.return_value = self._get_mock_method_resources(
            'nice_method')
        resource = Mock()
        attrs = utils.resource_view_attrs(resource, singular=False)
        assert attrs == set()

    def test_resource_schema_no_body(self):
        resource = Mock(body=None)
        with pytest.raises(ValueError) as ex:
            utils.resource_schema(resource)
        expected = 'RAML resource has no body to setup database'
        assert expected in str(ex.value)

    def test_resource_schema_no_schemas(self):
        resource = Mock(body=[Mock(schema=None), Mock(schema='')])
        assert utils.resource_schema(resource) is None

    def test_resource_schema_success(self):
        resource = Mock(body=[
            Mock(schema={'foo': 'bar'},
                 mime_type=utils.ContentTypes.JSON)
        ])
        assert utils.resource_schema(resource) == {'foo': 'bar'}

    def test_is_dynamic_resource_no_resource(self):
        assert not utils.is_dynamic_resource(None)

    def test_is_dynamic_resource_dynamic(self):
        resource = Mock(path='/{id}')
        assert utils.is_dynamic_resource(resource)

    def test_is_dynamic_resource_not_dynamic(self):
        resource = Mock(path='/stories')
        assert not utils.is_dynamic_resource(resource)

    def test_get_static_parent(self):
        parent = Mock(path='/stories', method='post')
        resource = Mock(path='/{id}')
        resource.parent = parent
        assert utils.get_static_parent(resource, method='post') is parent

    def test_get_static_parent_none(self):
        resource = Mock(path='/{id}')
        resource.parent = None
        assert utils.get_static_parent(resource, method='post') is None

    def test_get_static_parent_wrong_parent_method(self):
        root = Mock(resources=[
            Mock(path='/stories', method='options'),
            Mock(path='/users', method='post'),
            Mock(path='/stories', method='post'),
        ])
        parent = Mock(path='/stories', method='get', root=root)
        resource = Mock(path='/{id}')
        resource.parent = parent
        res = utils.get_static_parent(resource, method='post')
        assert res.method == 'post'
        assert res.path == '/stories'

    def test_get_static_parent_without_method_parent_present(self):
        root = Mock(resources=[
            Mock(path='/stories', method='options'),
            Mock(path='/stories', method='post'),
        ])
        parent = Mock(path='/stories', method='get', root=root)
        resource = Mock(path='/{id}')
        resource.parent = parent
        res = utils.get_static_parent(resource)
        assert res.method == 'get'
        assert res.path == '/stories'

    def test_get_static_parent_none_found_in_root(self):
        root = Mock(resources=[
            Mock(path='/stories', method='get'),
        ])
        parent = Mock(path='/stories', method='options', root=root)
        resource = Mock(path='/{id}')
        resource.parent = parent
        assert utils.get_static_parent(resource, method='post') is None

    @patch('ramses.utils.get_static_parent')
    @patch('ramses.utils.resource_schema')
    def test_attr_subresource_no_static_parent(self, mock_schema, mock_par):
        mock_par.return_value = None
        assert not utils.attr_subresource('foo', 1)
        mock_par.assert_called_once_with('foo', method='POST')

    @patch('ramses.utils.get_static_parent')
    @patch('ramses.utils.resource_schema')
    def test_attr_subresource_no_schema(self, mock_schema, mock_par):
        parent = Mock()
        mock_par.return_value = parent
        mock_schema.return_value = None
        assert not utils.attr_subresource('foo', 1)
        mock_par.assert_called_once_with('foo', method='POST')
        mock_schema.assert_called_once_with(parent)

    @patch('ramses.utils.get_static_parent')
    @patch('ramses.utils.resource_schema')
    def test_attr_subresource_not_attr(self, mock_schema, mock_par):
        parent = Mock()
        mock_par.return_value = parent
        mock_schema.return_value = {
            'properties': {
                'route_name': {
                    '_db_settings': {
                        'type': 'string'
                    }
                }
            }
        }
        assert not utils.attr_subresource('resource', 'route_name')
        mock_par.assert_called_once_with('resource', method='POST')
        mock_schema.assert_called_once_with(parent)

    @patch('ramses.utils.get_static_parent')
    @patch('ramses.utils.resource_schema')
    def test_attr_subresource_dict(self, mock_schema, mock_par):
        parent = Mock()
        mock_par.return_value = parent
        mock_schema.return_value = {
            'properties': {
                'route_name': {
                    '_db_settings': {
                        'type': 'dict'
                    }
                },
                'route_name2': {
                    '_db_settings': {
                        'type': 'list'
                    }
                }
            }
        }
        assert utils.attr_subresource('resource', 'route_name')
        mock_par.assert_called_once_with('resource', method='POST')
        mock_schema.assert_called_once_with(parent)
        assert utils.attr_subresource('resource', 'route_name2')

    @patch('ramses.utils.get_static_parent')
    @patch('ramses.utils.resource_schema')
    def test_singular_subresource_no_static_parent(self, mock_schema, mock_par):
        mock_par.return_value = None
        assert not utils.singular_subresource('foo', 1)
        mock_par.assert_called_once_with('foo', method='POST')

    @patch('ramses.utils.get_static_parent')
    @patch('ramses.utils.resource_schema')
    def test_singular_subresource_no_schema(self, mock_schema, mock_par):
        parent = Mock()
        mock_par.return_value = parent
        mock_schema.return_value = None
        assert not utils.singular_subresource('foo', 1)
        mock_par.assert_called_once_with('foo', method='POST')
        mock_schema.assert_called_once_with(parent)

    @patch('ramses.utils.get_static_parent')
    @patch('ramses.utils.resource_schema')
    def test_singular_subresource_not_attr(self, mock_schema, mock_par):
        parent = Mock()
        mock_par.return_value = parent
        mock_schema.return_value = {
            'properties': {
                'route_name': {
                    '_db_settings': {
                        'type': 'string'
                    }
                }
            }
        }
        assert not utils.singular_subresource('resource', 'route_name')
        mock_par.assert_called_once_with('resource', method='POST')
        mock_schema.assert_called_once_with(parent)

    @patch('ramses.utils.get_static_parent')
    @patch('ramses.utils.resource_schema')
    def test_singular_subresource_dict(self, mock_schema, mock_par):
        parent = Mock()
        mock_par.return_value = parent
        mock_schema.return_value = {
            'properties': {
                'route_name': {
                    '_db_settings': {
                        'type': 'relationship',
                        'uselist': False
                    }
                },
            }
        }
        assert utils.singular_subresource('resource', 'route_name')
        mock_par.assert_called_once_with('resource', method='POST')
        mock_schema.assert_called_once_with(parent)

    def test_is_callable_tag_not_str(self):
        assert not utils.is_callable_tag(1)
        assert not utils.is_callable_tag(None)

    def test_is_callable_tag_not_tag(self):
        assert not utils.is_callable_tag('foobar')

    def test_is_callable_tag(self):
        assert utils.is_callable_tag('{{foobar}}')

    def test_resolve_to_callable_not_found(self):
        with pytest.raises(ImportError) as ex:
            utils.resolve_to_callable('{{foobar}}')
        assert str(ex.value) == 'Failed to load callable `foobar`'

    def test_resolve_to_callable_registry(self):
        from ramses import registry

        @registry.add
        def foo():
            pass

        func = utils.resolve_to_callable('{{foo}}')
        assert func is foo
        func = utils.resolve_to_callable('foo')
        assert func is foo

    def test_resolve_to_callable_dotted_path(self):
        from datetime import datetime
        func = utils.resolve_to_callable('{{datetime.datetime}}')
        assert func is datetime
        func = utils.resolve_to_callable('datetime.datetime')
        assert func is datetime



import logging

import six
from pyramid.security import (
    Allow, Deny,
    Everyone, Authenticated,
    ALL_PERMISSIONS)
from nefertari.acl import CollectionACL
from nefertari.resource import PERMISSIONS
from nefertari.elasticsearch import ES

from .views import collection_methods, item_methods
from .utils import resolve_to_callable, is_callable_tag


log = logging.getLogger(__name__)


actions = {
    'allow': Allow,
    'deny': Deny,
}
special_principals = {
    'everyone': Everyone,
    'authenticated': Authenticated,
}
ALLOW_ALL = (Allow, Everyone, ALL_PERMISSIONS)


def methods_to_perms(perms, methods_map):
    """ Convert permissions ("perms") which are either HTTP methods or
    the keyword 'all' into a set of valid Pyramid permissions.

    :param perms: List or comma-separated string of HTTP methods, or 'all'
    :param methods_map: Map of HTTP methods to nefertari view methods
    """
    if isinstance(perms, six.string_types):
        perms = perms.split(',')
    perms = [perm.strip().lower() for perm in perms]
    if 'all' in perms:
        return ALL_PERMISSIONS
    else:
        try:
            return [PERMISSIONS[methods_map[p]] for p in perms]
        except KeyError:
            raise ValueError(
                'Unknown method name in permissions: {}. Valid methods: '
                '{}'.format(perms, list(methods_map.keys())))


def parse_acl(acl_string, methods_map):
    """ Parse raw string :acl_string: of RAML-defined ACLs.

    If :acl_string: is blank or None, all permissions are given.
    Values of ACL action and principal are parsed using `actions` and
    `special_principals` maps and are looked up after `strip()` and
    `lower()`.

    ACEs in :acl_string: may be separated by newlines or semicolons.
    Action, principal and permission lists must be separated by spaces.
    Permissions must be comma-separated.
    E.g. 'allow everyone get,post,patch' and 'deny authenticated delete'

    :param acl_string: Raw RAML string containing defined ACEs.
    :param methods_map: Map of HTTP methods to nefertari method handler names.
    """
    if not acl_string:
        return [ALLOW_ALL]

    aces_list = acl_string.replace('\n', ';').split(';')
    aces_list = [ace.strip().split(' ', 2) for ace in aces_list if ace]
    aces_list = [(a, b, c.split(',')) for a, b, c in aces_list]
    result_acl = []

    for action_str, princ_str, perms in aces_list:
        # Process action
        action_str = action_str.strip().lower()
        action = actions.get(action_str)
        if action is None:
            raise ValueError('Unknown ACL action: {}. Valid actions: {}'.format(
                action_str, list(actions.keys())))

        # Process principal
        princ_str = princ_str.strip().lower()
        if princ_str in special_principals:
            principal = special_principals[princ_str]
        elif is_callable_tag(princ_str):
            principal = resolve_to_callable(princ_str)
        else:
            principal = 'g:' + princ_str

        # Process permissions
        permissions = methods_to_perms(perms, methods_map)

        result_acl.append((action, principal, permissions))

    return result_acl


class BaseACL(CollectionACL):
    """ ACL Base class. """

    es_based = False
    _collection_acl = (ALLOW_ALL, )
    _item_acl = (ALLOW_ALL, )

    def _apply_callables(self, acl, methods_map, obj=None):
        """ Iterate over ACEs from :acl: and apply callable principals if any.

        Principals are passed 3 arguments on call:
            :ace: Single ACE object that looks like (action, callable,
                permission or [permission])
            :request: Current request object
            :obj: Object instance to be accessed via the ACL
        Principals must return a single ACE or a list of ACEs.

        :param acl: Sequence of valid Pyramid ACEs which will be processed
        :param methods_map: Map of HTTP methods to nefertari view method names
            (permissions)
        :param obj: Object to be accessed via the ACL
        """
        new_acl = []
        for i, ace in enumerate(acl):
            principal = ace[1]
            if six.callable(principal):
                ace = principal(ace=ace, request=self.request, obj=obj)
                if not ace:
                    continue
                if not isinstance(ace[0], (list, tuple)):
                    ace = [ace]
                ace = [(a, b, methods_to_perms(c, methods_map))
                       for a, b, c in ace]
            else:
                ace = [ace]
            new_acl += ace
        return new_acl

    def __acl__(self):
        """ Apply callables to `self._collection_acl` and return result. """
        return self._apply_callables(
            acl=self._collection_acl,
            methods_map=collection_methods)

    def item_acl(self, item):
        """ Apply callables to `self._item_acl` and return result. """
        return self._apply_callables(
            acl=self._item_acl,
            methods_map=item_methods,
            obj=item)

    def item_db_id(self, key):
        # ``self`` can be used for current authenticated user key
        if key != 'self':
            return key
        user = getattr(self.request, 'user', None)
        if user is None or not isinstance(user, self.item_model):
            return key
        return getattr(user, user.pk_field())

    def __getitem__(self, key):
        """ Get item using method depending on value of `self.es_based` """
        if not self.es_based:
            return super(BaseACL, self).__getitem__(key)
        return self.getitem_es(self.item_db_id(key))

    def getitem_es(self, key):
        es = ES(self.item_model.__name__)
        obj = es.get_resource(id=key)
        obj.__acl__ = self.item_acl(obj)
        obj.__parent__ = self
        obj.__name__ = key
        return obj


def generate_acl(model_cls, raml_resource, es_based=True):
    """ Generate an ACL.

    Generated ACL class has a `item_model` attribute set to
    :model_cls:.

    ACLs used for collection and item access control are generated from a
    first security scheme with type `x-ACL`.
    If :raml_resource: has no x-ACL security schemes defined then ALLOW_ALL
    ACL is used.
    If the `collection` or `item` settings are empty, then ALLOW_ALL ACL
    is used.

    :param model_cls: Generated model class
    :param raml_resource: Instance of ramlfications.raml.ResourceNode
        for which ACL is being generated
    :param es_based: Boolean inidicating whether ACL should query ES or
        not when getting an object
    """
    schemes = raml_resource.security_schemes or []
    schemes = [sch for sch in schemes if sch.type == 'x-ACL']

    if not schemes:
        collection_acl = item_acl = [ALLOW_ALL]
        log.debug('No ACL scheme applied. Using ACL: {}'.format(item_acl))
    else:
        sec_scheme = schemes[0]
        log.debug('{} ACL scheme applied'.format(sec_scheme.name))
        settings = sec_scheme.settings or {}
        collection_acl = parse_acl(
            acl_string=settings.get('collection'),
            methods_map=collection_methods)
        item_acl = parse_acl(
            acl_string=settings.get('item'),
            methods_map=item_methods)

    class GeneratedACL(BaseACL):
        item_model = model_cls

        def __init__(self, request, es_based=es_based):
            super(GeneratedACL, self).__init__(request=request)
            self.es_based = es_based
            self._collection_acl = collection_acl
            self._item_acl = item_acl

    return GeneratedACL



"""
Naive registry that is just a subclass of a python dictionary.
It is meant to be used to store objects and retrieve them when needed.
The registry is recreated on each app launch and is best suited to store some
dynamic or short-term data.

Storing an object should be performed by using the `add` function, and
retrieving it by using the `get` function.


Examples:

Register a function under a function name::

    from ramses import registry

    @registry.add
    def foo():
        print 'In foo'

    assert registry.get('foo') is foo


Register a function under a different name::

    from ramses import registry

    @registry.add('bar')
    def foo():
        print 'In foo'

    assert registry.get('bar') is foo


Register an arbitrary object::

    from ramses import registry

    myvar = 'my awesome var'
    registry.add('my_stored_var', myvar)
    assert registry.get('my_stored_var') == myvar


Register and get an object by namespace::

    from ramses import registry

    myvar = 'my awesome var'
    registry.add('Foo.my_stored_var', myvar)
    assert registry.mget('Foo') == {'my_stored_var': myvar}

"""
import six


class Registry(dict):
    pass


registry = Registry()


def add(*args):
    def decorator(function):
        registry[name] = function
        return function

    if len(args) == 1 and six.callable(args[0]):
        function = args[0]
        name = function.__name__
        return decorator(function)
    elif len(args) == 2:
        registry[args[0]] = args[1]
    else:
        name = args[0]
        return decorator


def get(name):
    try:
        return registry[name]
    except KeyError:
        raise KeyError(
            "Object named '{}' is not registered in ramses "
            "registry".format(name))


def mget(namespace):
    namespace = namespace.lower() + '.'
    data = {}
    for key, val in registry.items():
        key = key.lower()
        if not key.startswith(namespace):
            continue
        clean_key = key.split(namespace)[-1]
        data[clean_key] = val
    return data



import pytest
from mock import Mock, patch

from nefertari.utils import dictset

from .fixtures import engine_mock


@pytest.mark.usefixtures('engine_mock')
class TestSetupTicketPolicy(object):

    def test_no_secret(self):
        from ramses import auth
        with pytest.raises(ValueError) as ex:
            auth._setup_ticket_policy(config='', params={})
        expected = 'Missing required security scheme settings: secret'
        assert expected == str(ex.value)

    @patch('ramses.auth.AuthTktAuthenticationPolicy')
    def test_params_converted(self, mock_policy):
        from ramses import auth
        params = dictset(
            secure=True,
            include_ip=True,
            http_only=False,
            wild_domain=True,
            debug=True,
            parent_domain=True,
            secret='my_secret_setting'
        )
        auth_model = Mock()
        config = Mock()
        config.registry.settings = {'my_secret_setting': 12345}
        config.registry.auth_model = auth_model
        auth._setup_ticket_policy(config=config, params=params)
        mock_policy.assert_called_once_with(
            include_ip=True, secure=True, parent_domain=True,
            callback=auth_model.get_groups_by_userid, secret=12345,
            wild_domain=True, debug=True, http_only=False
        )

    @patch('ramses.auth.AuthTktAuthenticationPolicy')
    def test_request_method_added(self, mock_policy):
        from ramses import auth
        config = Mock()
        config.registry.settings = {'my_secret': 12345}
        config.registry.auth_model = Mock()
        policy = auth._setup_ticket_policy(
            config=config, params={'secret': 'my_secret'})
        config.add_request_method.assert_called_once_with(
            config.registry.auth_model.get_authuser_by_userid,
            'user', reify=True)
        assert policy == mock_policy()

    @patch('ramses.auth.AuthTktAuthenticationPolicy')
    def test_routes_views_added(self, mock_policy):
        from ramses import auth
        config = Mock()
        config.registry.settings = {'my_secret': 12345}
        config.registry.auth_model = Mock()
        root = Mock()
        config.get_root_resource.return_value = root
        auth._setup_ticket_policy(
            config=config, params={'secret': 'my_secret'})
        assert root.add.call_count == 3
        login, logout, register = root.add.call_args_list
        login_kwargs = login[1]
        assert sorted(login_kwargs.keys()) == sorted([
            'view', 'prefix', 'factory'])
        assert login_kwargs['prefix'] == 'auth'
        assert login_kwargs['factory'] == 'nefertari.acl.AuthenticationACL'

        logout_kwargs = logout[1]
        assert sorted(logout_kwargs.keys()) == sorted([
            'view', 'prefix', 'factory'])
        assert logout_kwargs['prefix'] == 'auth'
        assert logout_kwargs['factory'] == 'nefertari.acl.AuthenticationACL'

        register_kwargs = register[1]
        assert sorted(register_kwargs.keys()) == sorted([
            'view', 'prefix', 'factory'])
        assert register_kwargs['prefix'] == 'auth'
        assert register_kwargs['factory'] == 'nefertari.acl.AuthenticationACL'


@pytest.mark.usefixtures('engine_mock')
class TestSetupApiKeyPolicy(object):

    @patch('ramses.auth.ApiKeyAuthenticationPolicy')
    def test_policy_params(self, mock_policy):
        from ramses import auth
        auth_model = Mock()
        config = Mock()
        config.registry.auth_model = auth_model
        policy = auth._setup_apikey_policy(config, {'foo': 'bar'})
        mock_policy.assert_called_once_with(
            foo='bar', check=auth_model.get_groups_by_token,
            credentials_callback=auth_model.get_token_credentials,
            user_model=auth_model,
        )
        assert policy == mock_policy()

    @patch('ramses.auth.ApiKeyAuthenticationPolicy')
    def test_routes_views_added(self, mock_policy):
        from ramses import auth
        auth_model = Mock()
        config = Mock()
        config.registry.auth_model = auth_model
        root = Mock()
        config.get_root_resource.return_value = root
        auth._setup_apikey_policy(config, {})
        assert root.add.call_count == 3
        token, reset_token, register = root.add.call_args_list
        token_kwargs = token[1]
        assert sorted(token_kwargs.keys()) == sorted([
            'view', 'prefix', 'factory'])
        assert token_kwargs['prefix'] == 'auth'
        assert token_kwargs['factory'] == 'nefertari.acl.AuthenticationACL'

        reset_token_kwargs = reset_token[1]
        assert sorted(reset_token_kwargs.keys()) == sorted([
            'view', 'prefix', 'factory'])
        assert reset_token_kwargs['prefix'] == 'auth'
        assert reset_token_kwargs['factory'] == 'nefertari.acl.AuthenticationACL'

        register_kwargs = register[1]
        assert sorted(register_kwargs.keys()) == sorted([
            'view', 'prefix', 'factory'])
        assert register_kwargs['prefix'] == 'auth'
        assert register_kwargs['factory'] == 'nefertari.acl.AuthenticationACL'


@pytest.mark.usefixtures('engine_mock')
class TestSetupAuthPolicies(object):

    def test_not_secured(self):
        from ramses import auth
        raml_data = Mock(secured_by=[None])
        config = Mock()
        auth.setup_auth_policies(config, raml_data)
        assert not config.set_authentication_policy.called
        assert not config.set_authorization_policy.called

    def test_not_defined_security_scheme(self):
        from ramses import auth
        scheme = Mock()
        scheme.name = 'foo'
        raml_data = Mock(secured_by=['zoo'], security_schemes=[scheme])
        with pytest.raises(ValueError) as ex:
            auth.setup_auth_policies('asd', raml_data)
        expected = 'Undefined security scheme used in `secured_by`: zoo'
        assert expected == str(ex.value)

    def test_not_supported_scheme_type(self):
        from ramses import auth
        scheme = Mock(type='asd123')
        scheme.name = 'foo'
        raml_data = Mock(secured_by=['foo'], security_schemes=[scheme])
        with pytest.raises(ValueError) as ex:
            auth.setup_auth_policies(None, raml_data)
        expected = 'Unsupported security scheme type: asd123'
        assert expected == str(ex.value)

    @patch('ramses.auth.ACLAuthorizationPolicy')
    def test_policies_calls(self, mock_acl):
        from ramses import auth
        scheme = Mock(type='mytype', settings={'name': 'user1'})
        scheme.name = 'foo'
        raml_data = Mock(secured_by=['foo'], security_schemes=[scheme])
        config = Mock()
        mock_setup = Mock()
        with patch.dict(auth.AUTHENTICATION_POLICIES, {'mytype': mock_setup}):
            auth.setup_auth_policies(config, raml_data)
        mock_setup.assert_called_once_with(config, {'name': 'user1'})
        config.set_authentication_policy.assert_called_once_with(
            mock_setup())
        mock_acl.assert_called_once_with()
        config.set_authorization_policy.assert_called_once_with(
            mock_acl())


@pytest.mark.usefixtures('engine_mock')
class TestHelperFunctions(object):

    def test_create_system_user_key_error(self):
        from ramses import auth
        config = Mock()
        config.registry.settings = {}
        auth.create_system_user(config)
        assert not config.registry.auth_model.get_or_create.called

    @patch('ramses.auth.transaction')
    def test_create_system_user_exists(self, mock_trans):
        from ramses import auth
        config = Mock()
        config.registry.settings = {
            'system.user': 'user12',
            'system.password': '123456',
            'system.email': 'user12@example.com',
        }
        config.registry.auth_model.get_or_create.return_value = (1, False)
        auth.create_system_user(config)
        assert not mock_trans.commit.called
        config.registry.auth_model.get_or_create.assert_called_once_with(
            username='user12',
            defaults={
                'password': '123456',
                'email': 'user12@example.com',
                'groups': ['admin']
            }
        )

    @patch('ramses.auth.transaction')
    def test_create_system_user_created(self, mock_trans):
        from ramses import auth
        config = Mock()
        config.registry.settings = {
            'system.user': 'user12',
            'system.password': '123456',
            'system.email': 'user12@example.com',
        }
        config.registry.auth_model.get_or_create.return_value = (
            Mock(), True)
        auth.create_system_user(config)
        mock_trans.commit.assert_called_once_with()
        config.registry.auth_model.get_or_create.assert_called_once_with(
            username='user12',
            defaults={
                'password': '123456',
                'email': 'user12@example.com',
                'groups': ['admin']
            }
        )

    @patch('ramses.auth.create_system_user')
    def test_includeme(self, mock_create):
        from ramses import auth
        auth.includeme(config=1)
        mock_create.assert_called_once_with(1)



import logging

import six
from nefertari.view import BaseView as NefertariBaseView
from nefertari.json_httpexceptions import JHTTPNotFound


log = logging.getLogger(__name__)

"""
Maps of {HTTP_method: neferteri view method name}

"""
collection_methods = {
    'get':      'index',
    'head':     'index',
    'post':     'create',
    'put':      'update_many',
    'patch':    'update_many',
    'delete':   'delete_many',
    'options':  'collection_options',
}
item_methods = {
    'get':      'show',
    'head':     'show',
    'post':     'create',
    'put':      'replace',
    'patch':    'update',
    'delete':   'delete',
    'options':  'item_options',
}


class BaseView(NefertariBaseView):
    """ Base view class for other all views that defines few helper methods.

    Use `self.get_collection` and `self.get_item` to get access to set of
    objects and object respectively which are valid at current level.
    """
    @property
    def clean_id_name(self):
        id_name = self._resource.id_name
        if '_' in id_name:
            return id_name.split('_', 1)[1]
        else:
            return id_name

    def resolve_kw(self, kwargs):
        """ Resolve :kwargs: like `story_id: 1` to the form of `id: 1`.

        """
        resolved = {}
        for key, value in kwargs.items():
            split = key.split('_', 1)
            if len(split) > 1:
                key = split[1]
            resolved[key] = value
        return resolved

    def _location(self, obj):
        """ Get location of the `obj`

        Arguments:
            :obj: self.Model instance.
        """
        field_name = self.clean_id_name
        return self.request.route_url(
            self._resource.uid,
            **{self._resource.id_name: getattr(obj, field_name)})

    def _parent_queryset(self):
        """ Get queryset of parent view.

        Generated queryset is used to run queries in the current level view.
        """
        parent = self._resource.parent
        if hasattr(parent, 'view'):
            req = self.request.blank(self.request.path)
            req.registry = self.request.registry
            req.matchdict = {
                parent.id_name: self.request.matchdict.get(parent.id_name)}
            parent_view = parent.view(parent.view._factory, req)
            obj = parent_view.get_item(**req.matchdict)
            if isinstance(self, ItemSubresourceBaseView):
                return
            prop = self._resource.collection_name
            return getattr(obj, prop, None)

    def get_collection(self, **kwargs):
        """ Get objects collection taking into account generated queryset
        of parent view.

        This method allows working with nested resources properly. Thus a
        queryset returned by this method will be a subset of its parent
        view's queryset, thus filtering out objects that don't belong to
        the parent object.
        """
        self._query_params.update(kwargs)
        objects = self._parent_queryset()
        if objects is not None:
            return self.Model.filter_objects(
                objects, **self._query_params)
        return self.Model.get_collection(**self._query_params)

    def get_item(self, **kwargs):
        """ Get collection item taking into account generated queryset
        of parent view.

        This method allows working with nested resources properly. Thus an item
        returned by this method will belong to its parent view's queryset, thus
        filtering out objects that don't belong to the parent object.

        Returns an object from the applicable ACL. If ACL wasn't applied, it is
        applied explicitly.
        """
        if six.callable(self.context):
            self.reload_context(es_based=False, **kwargs)

        objects = self._parent_queryset()
        if objects is not None and self.context not in objects:
            raise JHTTPNotFound('{}({}) not found'.format(
                self.Model.__name__,
                self._get_context_key(**kwargs)))

        return self.context

    def _get_context_key(self, **kwargs):
        """ Get value of `self._resource.id_name` from :kwargs: """
        return str(kwargs.get(self._resource.id_name))

    def reload_context(self, es_based, **kwargs):
        """ Reload `self.context` object into a DB or ES object.

        A reload is performed by getting the object ID from :kwargs: and then
        getting a context key item from the new instance of `self._factory`
        which is an ACL class used by the current view.

        Arguments:
            :es_based: Boolean. Whether to init ACL ac es-based or not. This
                affects the backend which will be queried - either DB or ES
            :kwargs: Kwargs that contain value for current resource 'id_name'
                key
        """
        from .acl import BaseACL
        key = self._get_context_key(**kwargs)
        kwargs = {'request': self.request}
        if issubclass(self._factory, BaseACL):
            kwargs['es_based'] = es_based

        acl = self._factory(**kwargs)
        if acl.item_model is None:
            acl.item_model = self.Model

        self.context = acl[key]


class CollectionView(BaseView):
    """ View that works with database and implements handlers for all
    available CRUD operations.

    """
    def index(self, **kwargs):
        return self.get_collection()

    def show(self, **kwargs):
        return self.get_item(**kwargs)

    def create(self, **kwargs):
        obj = self.Model(**self._json_params)
        return obj.save(self.request)

    def update(self, **kwargs):
        obj = self.get_item(**kwargs)
        return obj.update(self._json_params, self.request)

    def replace(self, **kwargs):
        return self.update(**kwargs)

    def delete(self, **kwargs):
        obj = self.get_item(**kwargs)
        obj.delete(self.request)

    def delete_many(self, **kwargs):
        objects = self.get_collection()

        if self.needs_confirmation():
            return objects

        return self.Model._delete_many(objects, self.request)

    def update_many(self, **kwargs):
        objects = self.get_collection(**self._query_params)
        return self.Model._update_many(
            objects, self._json_params, self.request)


class ESBaseView(BaseView):
    """ Elasticsearch base view that fetches data from ES.

    Implements analogues of _parent_queryset, get_collection, get_item
    fetching data from ES instead of database.

    Use `self.get_collection_es` and `self.get_item_es` to get access
    to the set of objects and individual object respectively which are
    valid at the current level.
    """
    def _parent_queryset_es(self):
        """ Get queryset (list of object IDs) of parent view.

        The generated queryset is used to run queries in the current level's
        view.
        """
        parent = self._resource.parent
        if hasattr(parent, 'view'):
            req = self.request.blank(self.request.path)
            req.registry = self.request.registry
            req.matchdict = {
                parent.id_name: self.request.matchdict.get(parent.id_name)}
            parent_view = parent.view(parent.view._factory, req)
            obj = parent_view.get_item_es(**req.matchdict)
            prop = self._resource.collection_name
            objects_ids = getattr(obj, prop, None)
            return objects_ids

    def get_es_object_ids(self, objects):
        """ Return IDs of :objects: if they are not IDs already. """
        id_field = self.clean_id_name
        ids = [getattr(obj, id_field, obj) for obj in objects]
        return list(set(str(id_) for id_ in ids))

    def get_collection_es(self, **kwargs):
        """ Get ES objects collection taking into account the generated
        queryset of parent view.

        This method allows working with nested resources properly. Thus a
        queryset returned by this method will be a subset of its parent view's
        queryset, thus filtering out objects that don't belong to the parent
        object.
        """
        from nefertari.elasticsearch import ES
        es = ES(self.Model.__name__)
        objects_ids = self._parent_queryset_es()

        if objects_ids is not None:
            objects_ids = self.get_es_object_ids(objects_ids)

            if not objects_ids:
                return []
            self._query_params['id'] = objects_ids
        return es.get_collection(**self._query_params)

    def get_item_es(self, **kwargs):
        """ Get ES collection item taking into account generated queryset
        of parent view.

        This method allows working with nested resources properly. Thus an item
        returned by this method will belong to its parent view's queryset, thus
        filtering out objects that don't belong to the parent object.

        Returns an object retrieved from the applicable ACL. If an ACL wasn't
        applied, it is applied explicitly.
        """
        item_id = self._get_context_key(**kwargs)
        objects_ids = self._parent_queryset_es()
        if objects_ids is not None:
            objects_ids = self.get_es_object_ids(objects_ids)

        if six.callable(self.context):
            self.reload_context(es_based=True, **kwargs)

        if (objects_ids is not None) and (item_id not in objects_ids):
            raise JHTTPNotFound('{}(id={}) resource not found'.format(
                self.Model.__name__, item_id))

        return self.context


class ESCollectionView(ESBaseView, CollectionView):
    """ View that reads data from ES.

    Write operations are inherited from :CollectionView:
    """
    def index(self, **kwargs):
        return self.get_collection_es(**kwargs)

    def show(self, **kwargs):
        return self.get_item_es(**kwargs)

    def update(self, **kwargs):
        """ Explicitly reload context with DB usage to get access
        to complete DB object.
        """
        self.reload_context(es_based=False, **kwargs)
        return super(ESCollectionView, self).update(**kwargs)

    def delete(self, **kwargs):
        """ Explicitly reload context with DB usage to get access
        to complete DB object.
        """
        self.reload_context(es_based=False, **kwargs)
        return super(ESCollectionView, self).delete(**kwargs)

    def get_dbcollection_with_es(self, **kwargs):
        """ Get DB objects collection by first querying ES. """
        es_objects = self.get_collection_es(**kwargs)
        db_objects = self.Model.filter_objects(es_objects)
        return db_objects

    def delete_many(self, **kwargs):
        """ Delete multiple objects from collection.

        First ES is queried, then the results are used to query the DB.
        This is done to make sure deleted objects are those filtered
        by ES in the 'index' method (so user deletes what he saw).
        """
        db_objects = self.get_dbcollection_with_es(**kwargs)

        if self.needs_confirmation():
            return db_objects

        return self.Model._delete_many(db_objects, self.request)

    def update_many(self, **kwargs):
        """ Update multiple objects from collection.

        First ES is queried, then the results are used to query DB.
        This is done to make sure updated objects are those filtered
        by ES in the 'index' method (so user updates what he saw).
        """
        db_objects = self.get_dbcollection_with_es(**kwargs)
        return self.Model._update_many(
            db_objects, self._json_params, self.request)


class ItemSubresourceBaseView(BaseView):
    """ Base class for all subresources of collection item resources which
    don't represent a collection of their own.
    E.g. /users/{id}/profile, where 'profile' is a singular resource or
    /users/{id}/some_action, where the 'some_action' action may be
    performed when requesting this route.

    Subclass ItemSubresourceBaseView in your project when you want to
    define a subroute and view of an item route defined in RAML and
    generated by ramses.
    Use `self.get_item` to get an object on which actions are being
    performed.

    Moved into a separate class so all item subresources have a common
    base class, thus making checks like `isinstance(view, baseClass)` easier.
    Also to override `_get_context_key` to return parent resource's id_name
    and `get_item` to reload context on each access.
    """

    def _get_context_key(self, **kwargs):
        """ Get value of `self._resource.parent.id_name` from :kwargs: """
        return str(kwargs.get(self._resource.parent.id_name))

    def get_item(self, **kwargs):
        """ Reload context on each access. """
        self.reload_context(es_based=False, **kwargs)
        return super(ItemSubresourceBaseView, self).get_item(**kwargs)


class ItemAttributeView(ItemSubresourceBaseView):
    """ View used to work with attribute resources.

    Attribute resources represent field: ListField, DictField.

    You may subclass ItemAttributeView in your project when you want to
    define custom attribute subroute and view of a item route defined in
    RAML and generated by ramses.
    """
    def __init__(self, *args, **kw):
        super(ItemAttributeView, self).__init__(*args, **kw)
        self.attr = self.request.path.split('/')[-1]
        self.value_type = None
        self.unique = True

    def index(self, **kwargs):
        obj = self.get_item(**kwargs)
        return getattr(obj, self.attr)

    def create(self, **kwargs):
        obj = self.get_item(**kwargs)
        obj.update_iterables(
            self._json_params, self.attr,
            unique=self.unique,
            value_type=self.value_type,
            request=self.request)
        return getattr(obj, self.attr, None)


class ItemSingularView(ItemSubresourceBaseView):
    """ View used to work with singular resources.

    Singular resources represent a one-to-one relationship.
    E.g. users/1/profile.

    You may subclass ItemSingularView in your project when you want to define
    a custom singular subroute and view of an item route defined in RAML and
    generated by ramses.
    If you decide to do so, make sure to set `self._singular_model` to a model
    class, instances of which will be processed by this view.
    """
    _singular_model = None

    def __init__(self, *args, **kw):
        super(ItemSingularView, self).__init__(*args, **kw)
        self.attr = self.request.path.split('/')[-1]

    def fill_null_values(self, model_cls=None):
        return super(ItemSingularView, self).fill_null_values(
            model_cls=self._singular_model)

    def convert_ids2objects(self, model_cls=None):
        return super(ItemSingularView, self).convert_ids2objects(
            model_cls=self._singular_model)

    def show(self, **kwargs):
        parent_obj = self.get_item(**kwargs)
        return getattr(parent_obj, self.attr)

    def create(self, **kwargs):
        parent_obj = self.get_item(**kwargs)
        obj = self._singular_model(**self._json_params)
        obj = obj.save(self.request)
        parent_obj.update({self.attr: obj}, self.request)
        return obj

    def update(self, **kwargs):
        parent_obj = self.get_item(**kwargs)
        obj = getattr(parent_obj, self.attr)
        obj.update(self._json_params, self.request)
        return obj

    def replace(self, **kwargs):
        return self.update(**kwargs)

    def delete(self, **kwargs):
        parent_obj = self.get_item(**kwargs)
        obj = getattr(parent_obj, self.attr)
        obj.delete(self.request)


def generate_rest_view(model_cls, attrs=None, es_based=True,
                       attr_view=False, singular=False):
    """ Generate REST view for a model class.

    :param model_cls: Generated DB model class.
    :param attr: List of strings that represent names of view methods, new
        generated view should support. Not supported methods are replaced
        with property that raises AttributeError to display MethodNotAllowed
        error.
    :param es_based: Boolean indicating if generated view should read from
        elasticsearch. If True - collection reads are performed from
        elasticsearch. Database is used for reads otherwise.
        Defaults to True.
    :param attr_view: Boolean indicating if ItemAttributeView should be
        used as a base class for generated view.
    :param singular: Boolean indicating if ItemSingularView should be
        used as a base class for generated view.
    """
    valid_attrs = (list(collection_methods.values()) +
                   list(item_methods.values()))
    missing_attrs = set(valid_attrs) - set(attrs)

    if singular:
        base_view_cls = ItemSingularView
    elif attr_view:
        base_view_cls = ItemAttributeView
    elif es_based:
        base_view_cls = ESCollectionView
    else:
        base_view_cls = CollectionView

    def _attr_error(*args, **kwargs):
        raise AttributeError

    class RESTView(base_view_cls):
        Model = model_cls

    for attr in missing_attrs:
        setattr(RESTView, attr, property(_attr_error))

    return RESTView



"""
Auth module that contains all code needed for authentication/authorization
policies setup.

In particular:
    :includeme: Function that actually creates routes listed above and
        connects view to them
    :create_system_user: Function that creates system/admin user
    :_setup_ticket_policy: Setup Pyramid AuthTktAuthenticationPolicy
    :_setup_apikey_policy: Setup nefertari.ApiKeyAuthenticationPolicy
    :setup_auth_policies: Runs generation of particular auth policy
"""
import logging

import transaction
from pyramid.authentication import AuthTktAuthenticationPolicy
from pyramid.authorization import ACLAuthorizationPolicy

from nefertari.utils import dictset
from nefertari.json_httpexceptions import *
from nefertari.authentication.policies import ApiKeyAuthenticationPolicy

log = logging.getLogger(__name__)


def _setup_ticket_policy(config, params):
    """ Setup Pyramid AuthTktAuthenticationPolicy.

    Notes:
      * Initial `secret` params value is considered to be a name of config
        param that represents a cookie name.
      * `auth_model.get_groups_by_userid` is used as a `callback`.
      * Also connects basic routes to perform authentication actions.

    :param config: Pyramid Configurator instance.
    :param params: Nefertari dictset which contains security scheme
        `settings`.
    """
    from nefertari.authentication.views import (
        TicketAuthRegisterView, TicketAuthLoginView,
        TicketAuthLogoutView)

    log.info('Configuring Pyramid Ticket Authn policy')
    if 'secret' not in params:
        raise ValueError(
            'Missing required security scheme settings: secret')
    params['secret'] = config.registry.settings[params['secret']]

    auth_model = config.registry.auth_model
    params['callback'] = auth_model.get_groups_by_userid

    config.add_request_method(
        auth_model.get_authuser_by_userid, 'user', reify=True)

    policy = AuthTktAuthenticationPolicy(**params)

    class RamsesTicketAuthRegisterView(TicketAuthRegisterView):
        Model = config.registry.auth_model

    class RamsesTicketAuthLoginView(TicketAuthLoginView):
        Model = config.registry.auth_model

    class RamsesTicketAuthLogoutView(TicketAuthLogoutView):
        Model = config.registry.auth_model

    common_kw = {
        'prefix': 'auth',
        'factory': 'nefertari.acl.AuthenticationACL',
    }

    root = config.get_root_resource()
    root.add('register', view=RamsesTicketAuthRegisterView, **common_kw)
    root.add('login', view=RamsesTicketAuthLoginView, **common_kw)
    root.add('logout', view=RamsesTicketAuthLogoutView, **common_kw)

    return policy


def _setup_apikey_policy(config, params):
    """ Setup `nefertari.ApiKeyAuthenticationPolicy`.

    Notes:
      * User may provide model name in :params['user_model']: do define
        the name of the user model.
      * `auth_model.get_groups_by_token` is used to perform username and
        token check
      * `auth_model.get_token_credentials` is used to get username and
        token from userid
      * Also connects basic routes to perform authentication actions.

    Arguments:
        :config: Pyramid Configurator instance.
        :params: Nefertari dictset which contains security scheme `settings`.
    """
    from nefertari.authentication.views import (
        TokenAuthRegisterView, TokenAuthClaimView,
        TokenAuthResetView)
    log.info('Configuring ApiKey Authn policy')

    auth_model = config.registry.auth_model
    params['check'] = auth_model.get_groups_by_token
    params['credentials_callback'] = auth_model.get_token_credentials
    params['user_model'] = auth_model
    config.add_request_method(
        auth_model.get_authuser_by_name, 'user', reify=True)

    policy = ApiKeyAuthenticationPolicy(**params)

    class RamsesTokenAuthRegisterView(TokenAuthRegisterView):
        Model = auth_model

    class RamsesTokenAuthClaimView(TokenAuthClaimView):
        Model = auth_model

    class RamsesTokenAuthResetView(TokenAuthResetView):
        Model = auth_model

    common_kw = {
        'prefix': 'auth',
        'factory': 'nefertari.acl.AuthenticationACL',
    }

    root = config.get_root_resource()
    root.add('register', view=RamsesTokenAuthRegisterView, **common_kw)
    root.add('token', view=RamsesTokenAuthClaimView, **common_kw)
    root.add('reset_token', view=RamsesTokenAuthResetView, **common_kw)

    return policy


""" Map of `security_scheme_type`: `generator_function`, where:

  * `security_scheme_type`: String that represents RAML security scheme type
    name that should be used to apply a particular authentication system.
  * `generator_function`: Function that receives instance of Pyramid
    Configurator instance and dictset of security scheme settings and returns
    generated Pyramid authentication policy instance.

"""
AUTHENTICATION_POLICIES = {
    'x-ApiKey': _setup_apikey_policy,
    'x-Ticket': _setup_ticket_policy,
}


def setup_auth_policies(config, raml_root):
    """ Setup authentication, authorization policies.

    Performs basic validation to check all the required values are present
    and performs authentication, authorization policies generation using
    generator functions from `AUTHENTICATION_POLICIES`.

    :param config: Pyramid Configurator instance.
    :param raml_root: Instance of ramlfications.raml.RootNode.
    """
    log.info('Configuring auth policies')
    secured_by_all = raml_root.secured_by or []
    secured_by = [item for item in secured_by_all if item]
    if not secured_by:
        log.info('API is not secured. `secured_by` attribute '
                 'value missing.')
        return
    secured_by = secured_by[0]

    schemes = {scheme.name: scheme
               for scheme in raml_root.security_schemes}
    if secured_by not in schemes:
        raise ValueError(
            'Undefined security scheme used in `secured_by`: {}'.format(
                secured_by))

    scheme = schemes[secured_by]
    if scheme.type not in AUTHENTICATION_POLICIES:
        raise ValueError('Unsupported security scheme type: {}'.format(
            scheme.type))

    # Setup Authentication policy
    policy_generator = AUTHENTICATION_POLICIES[scheme.type]
    params = dictset(scheme.settings or {})
    authn_policy = policy_generator(config, params)
    config.set_authentication_policy(authn_policy)

    # Setup Authorization policy
    authz_policy = ACLAuthorizationPolicy()
    config.set_authorization_policy(authz_policy)


def create_system_user(config):
    log.info('Creating system user')
    settings = config.registry.settings
    try:
        auth_model = config.registry.auth_model
        s_user = settings['system.user']
        s_pass = settings['system.password']
        s_email = settings['system.email']
        user, created = auth_model.get_or_create(
            username=s_user,
            defaults=dict(
                password=s_pass,
                email=s_email,
                groups=['admin']
            ))
        if created:
            transaction.commit()
    except KeyError as e:
        log.error('Failed to create system user. Missing config: %s' % e)


def get_authuser_model():
    """ Define and return AuthUser model using nefertari base classes """
    from nefertari.authentication.models import AuthUserMixin
    from nefertari import engine

    class AuthUser(AuthUserMixin, engine.BaseDocument):
        __tablename__ = 'ramses_authuser'

    return AuthUser


def includeme(config):
    create_system_user(config)



import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.md')).read()
VERSION = open(os.path.join(here, 'VERSION')).read()

requires = [
    'pyramid',
    'cryptacular',
    'ramlfications==0.1.5',
    'inflection',
    'transaction',
    'six',
    'nefertari>=0.5.1',
]

setup(name='ramses',
      version=VERSION,
      description='Generate a RESTful API for Pyramid using RAML',
      long_description=README,
      classifiers=[
          "Programming Language :: Python",
          "Programming Language :: Python :: 2",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.4",
          "Framework :: Pyramid",
          "Topic :: Internet :: WWW/HTTP",
          "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
      ],
      author='Brandicted',
      author_email='hello@brandicted.com',
      url='https://github.com/brandicted/ramses',
      keywords='web pyramid pylons nefertari raml',
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      install_requires=requires,
      tests_require=requires,
      test_suite="ramses",
      entry_points="""\
        [pyramid.scaffold]
        ramses_starter = ramses.scaffolds:RamsesStarterTemplate
      """)



import logging

import six
import inflection


log = logging.getLogger(__name__)


class ContentTypes(object):
    """ ContentType values.

    """
    JSON = 'application/json'
    TEXT_XML = 'text/xml'
    MULTIPART_FORMDATA = 'multipart/form-data'
    FORM_URLENCODED = 'application/x-www-form-urlencoded'


def convert_schema(raml_schema, mime_type):
    """ Restructure `raml_schema` to a dictionary that has 'properties'
    as well as other schema keys/values.

    The resulting dictionary looks like this::

    {
        "properties": {
            "field1": {
                "required": boolean,
                "type": ...,
                ...more field options
            },
            ...more properties
        },
        "public_fields": [...],
        "auth_fields": [...],
        ...more schema options
    }

    :param raml_schema: RAML request body schema.
    :param mime_type: ContentType of the schema as a string from RAML
        file. Only JSON is currently supported.
    """
    if mime_type == ContentTypes.JSON:
        if not isinstance(raml_schema, dict):
            raise TypeError(
                'Schema is not a valid JSON. Please check your '
                'schema syntax.\n{}...'.format(str(raml_schema)[:60]))
        return raml_schema
    if mime_type == ContentTypes.TEXT_XML:
        # Process XML schema
        pass


def is_dynamic_uri(uri):
    """ Determine whether `uri` is a dynamic uri or not.

    Assumes a dynamic uri is one that ends with '}' which is a Pyramid
    way to define dynamic parts in uri.

    :param uri: URI as a string.
    """
    return uri.strip('/').endswith('}')


def clean_dynamic_uri(uri):
    """ Strips /, {, } from dynamic `uri`.

    :param uri: URI as a string.
    """
    return uri.replace('/', '').replace('{', '').replace('}', '')


def generate_model_name(name):
    """ Generate model name.

    :param name: String representing a field or route name.
    """
    model_name = inflection.camelize(name.strip('/'))
    return inflection.singularize(model_name)


def dynamic_part_name(raml_resource, clean_uri, pk_field):
    """ Generate a dynamic part for a resource :raml_resource:.

    A dynamic part is generated using 2 parts: :clean_uri: of the resource
    and the dynamic part of first dynamic child resources. If
    :raml_resource: has no dynamic child resources, 'id' is used as the
    2nd part.
    E.g. if your dynamic part on route 'stories' is named 'superId' then
    dynamic part will be 'stories_superId'.

    :param raml_resource: Instance of ramlfications.raml.ResourceNode for
        which dynamic part name is being generated.
    :param clean_uri: Cleaned URI of :raml_resource:
    :param pk_field: Model Primary Key field name.
    """
    subresources = get_resource_children(raml_resource)
    dynamic_uris = [res.path for res in subresources
                    if is_dynamic_uri(res.path)]
    if dynamic_uris:
        dynamic_part = extract_dynamic_part(dynamic_uris[0])
    else:
        dynamic_part = pk_field
    return '_'.join([clean_uri, dynamic_part])


def extract_dynamic_part(uri):
    """ Extract dynamic url part from :uri: string.

    :param uri: URI string that may contain dynamic part.
    """
    for part in uri.split('/'):
        part = part.strip()
        if part.startswith('{') and part.endswith('}'):
            return clean_dynamic_uri(part)


def resource_view_attrs(raml_resource, singular=False):
    """ Generate view method names needed for `raml_resource` view.

    Collects HTTP method names from resource siblings and dynamic children
    if exist. Collected methods are then translated  to
    `nefertari.view.BaseView` method names, each of which is used to
    process a particular HTTP method request.

    Maps of {HTTP_method: view_method} `collection_methods` and
    `item_methods` are used to convert collection and item methods
    respectively.

    :param raml_resource: Instance of ramlfications.raml.ResourceNode
    :param singular: Boolean indicating if resource is singular or not
    """
    from .views import collection_methods, item_methods
    # Singular resource doesn't have collection methods though
    # it looks like a collection
    if singular:
        collection_methods = item_methods

    siblings = get_resource_siblings(raml_resource)
    http_methods = [sibl.method.lower() for sibl in siblings]
    attrs = [collection_methods.get(method) for method in http_methods]

    # Check if resource has dynamic child resource like collection/{id}
    # If dynamic child resource exists, add its siblings' methods to attrs,
    # as both resources are handled by a single view
    children = get_resource_children(raml_resource)
    http_submethods = [child.method.lower() for child in children
                       if is_dynamic_uri(child.path)]
    attrs += [item_methods.get(method) for method in http_submethods]

    return set(filter(bool, attrs))


def resource_schema(raml_resource):
    """ Get schema properties of RAML resource :raml_resource:.

    Must be called with RAML resource that defines body schema. First
    body that defines schema is used. Schema is converted on return using
    'convert_schema'.

    :param raml_resource: Instance of ramlfications.raml.ResourceNode of
        POST method.
    """
    # NOTE: Must be called with resource that defines body schema
    log.info('Searching for model schema')
    if not raml_resource.body:
        raise ValueError('RAML resource has no body to setup database '
                         'schema from')

    for body in raml_resource.body:
        if body.schema:
            return convert_schema(body.schema, body.mime_type)
    log.debug('No model schema found.')


def is_dynamic_resource(raml_resource):
    """ Determine if :raml_resource: is a dynamic resource.

    :param raml_resource: Instance of ramlfications.raml.ResourceNode.
    """
    return raml_resource and is_dynamic_uri(raml_resource.path)


def get_static_parent(raml_resource, method=None):
    """ Get static parent resource of :raml_resource: with HTTP
    method :method:.

    :param raml_resource:Instance of ramlfications.raml.ResourceNode.
    :param method: HTTP method name which matching static resource
        must have.
    """
    parent = raml_resource.parent
    while is_dynamic_resource(parent):
        parent = parent.parent

    if parent is None:
        return parent

    match_method = method is not None
    if match_method:
        if parent.method.upper() == method.upper():
            return parent
    else:
        return parent

    for res in parent.root.resources:
        if res.path == parent.path:
            if res.method.upper() == method.upper():
                return res


def attr_subresource(raml_resource, route_name):
    """ Determine if :raml_resource: is an attribute subresource.

    :param raml_resource: Instance of ramlfications.raml.ResourceNode.
    :param route_name: Name of the :raml_resource:.
    """
    static_parent = get_static_parent(raml_resource, method='POST')
    if static_parent is None:
        return False
    schema = resource_schema(static_parent) or {}
    properties = schema.get('properties', {})
    if route_name in properties:
        db_settings = properties[route_name].get('_db_settings', {})
        return db_settings.get('type') in ('dict', 'list')
    return False


def singular_subresource(raml_resource, route_name):
    """ Determine if :raml_resource: is a singular subresource.

    :param raml_resource: Instance of ramlfications.raml.ResourceNode.
    :param route_name: Name of the :raml_resource:.
    """
    static_parent = get_static_parent(raml_resource, method='POST')
    if static_parent is None:
        return False
    schema = resource_schema(static_parent) or {}
    properties = schema.get('properties', {})
    if route_name not in properties:
        return False

    db_settings = properties[route_name].get('_db_settings', {})
    is_obj = db_settings.get('type') == 'relationship'
    single_obj = not db_settings.get('uselist', True)
    return is_obj and single_obj


def is_callable_tag(tag):
    """ Determine whether :tag: is a valid callable string tag.

    String is assumed to be valid callable if it starts with '{{'
    and ends with '}}'.

    :param tag: String name of tag.
    """
    return (isinstance(tag, six.string_types) and
            tag.strip().startswith('{{') and
            tag.strip().endswith('}}'))


def resolve_to_callable(callable_name):
    """ Resolve string :callable_name: to a callable.

    :param callable_name: String representing callable name as registered
        in ramses registry or dotted import path of callable. Can be
        wrapped in double curly brackets, e.g. '{{my_callable}}'.
    """
    from . import registry
    clean_callable_name = callable_name.replace(
        '{{', '').replace('}}', '').strip()
    try:
        return registry.get(clean_callable_name)
    except KeyError:
        try:
            from zope.dottedname.resolve import resolve
            return resolve(clean_callable_name)
        except ImportError:
            raise ImportError(
                'Failed to load callable `{}`'.format(clean_callable_name))


def get_resource_siblings(raml_resource):
    """ Get siblings of :raml_resource:.

    :param raml_resource: Instance of ramlfications.raml.ResourceNode.
    """
    path = raml_resource.path
    return [res for res in raml_resource.root.resources
            if res.path == path]


def get_resource_children(raml_resource):
    """ Get children of :raml_resource:.

    :param raml_resource: Instance of ramlfications.raml.ResourceNode.
    """
    path = raml_resource.path
    return [res for res in raml_resource.root.resources
            if res.parent and res.parent.path == path]



import binascii
import os
import subprocess

from six import moves
from pyramid.scaffolds import PyramidTemplate


class RamsesStarterTemplate(PyramidTemplate):
    _template_dir = 'ramses_starter'
    summary = 'Ramses starter'

    def pre(self, command, output_dir, vars):
        dbengine_choices = {'1': 'sqla', '2': 'mongodb'}
        vars['engine'] = dbengine_choices[moves.input("""
        Which database backend would you like to use:

        (1) for SQLAlchemy/PostgreSQL, or
        (2) for MongoEngine/MongoDB?

        [default is '1']: """) or '1']

        if vars['package'] == 'site':
            raise ValueError("""
                "Site" is a reserved keyword in Python.
                 Please use a different project name. """)

    def post(self, command, output_dir, vars):
        os.chdir(str(output_dir))
        subprocess.call('pip install -r requirements.txt', shell=True)
        subprocess.call('pip install nefertari-{}'.format(vars['engine']),
                        shell=True)
        msg = """Goodbye boilerplate! Welcome to Ramses."""
        self.out(msg)



import logging

import ramlfications
from nefertari.acl import RootACL as NefertariRootACL
from nefertari.utils import dictset


log = logging.getLogger(__name__)


def includeme(config):
    from .generators import generate_server, generate_models
    Settings = dictset(config.registry.settings)
    config.include('nefertari.engine')
    config.include('nefertari')
    config.include('nefertari.view')
    config.include('nefertari.json_httpexceptions')

    # Process nefertari settings
    if Settings.asbool('enable_get_tunneling'):
        config.add_tween('nefertari.tweens.get_tunneling')

    if Settings.asbool('cors.enable'):
        config.add_tween('nefertari.tweens.cors')

    if Settings.asbool('ssl_middleware.enable'):
        config.add_tween('nefertari.tweens.ssl')

    if Settings.asbool('request_timing.enable'):
        config.add_tween('nefertari.tweens.request_timing')

    # Set root factory
    config.root_factory = NefertariRootACL

    # Process auth settings
    root = config.get_root_resource()
    root_auth = getattr(root, 'auth', False)

    log.info('Parsing RAML')
    raml_root = ramlfications.parse(Settings['ramses.raml_schema'])

    log.info('Starting models generation')
    generate_models(config, raml_resources=raml_root.resources)

    if root_auth:
        from .auth import setup_auth_policies, get_authuser_model
        if getattr(config.registry, 'auth_model', None) is None:
            config.registry.auth_model = get_authuser_model()
        setup_auth_policies(config, raml_root)

    config.include('nefertari.elasticsearch')

    log.info('Starting server generation')
    generate_server(raml_root, config)

    log.info('Running nefertari.engine.setup_database')
    from nefertari.engine import setup_database
    setup_database(config)

    from nefertari.elasticsearch import ES
    ES.setup_mappings()

    if root_auth:
        config.include('ramses.auth')

    log.info('Server succesfully generated\n')



import pytest
from mock import Mock, patch, call

from .fixtures import engine_mock


@pytest.mark.usefixtures('engine_mock')
class TestHelperFunctions(object):

    @patch('ramses.models.engine')
    def test_get_existing_model_not_found(self, mock_eng):
        from ramses import models
        mock_eng.get_document_cls.side_effect = ValueError
        model_cls = models.get_existing_model('Foo')
        assert model_cls is None
        mock_eng.get_document_cls.assert_called_once_with('Foo')

    @patch('ramses.models.engine')
    def test_get_existing_model_found(self, mock_eng):
        from ramses import models
        mock_eng.get_document_cls.return_value = 1
        model_cls = models.get_existing_model('Foo')
        assert model_cls == 1
        mock_eng.get_document_cls.assert_called_once_with('Foo')

    @patch('ramses.models.setup_data_model')
    @patch('ramses.models.get_existing_model')
    def test_prepare_relationship_model_exists(self, mock_get, mock_set):
        from ramses import models
        models.prepare_relationship('foobar', 'Story', 'raml_resource')
        mock_get.assert_called_once_with('Story')
        assert not mock_set.called

    @patch('ramses.models.get_existing_model')
    def test_prepare_relationship_resource_not_found(self, mock_get):
        from ramses import models
        resource = Mock(root=Mock(resources=[
            Mock(method='get', path='/stories'),
            Mock(method='options', path='/stories'),
            Mock(method='post', path='/items'),
        ]))
        mock_get.return_value = None
        with pytest.raises(ValueError) as ex:
            models.prepare_relationship('stories', 'Story', resource)
        expected = ('Model `Story` used in relationship `stories` '
                    'is not defined')
        assert str(ex.value) == expected

    @patch('ramses.models.setup_data_model')
    @patch('ramses.models.get_existing_model')
    def test_prepare_relationship_resource_found(
            self, mock_get, mock_set):
        from ramses import models
        matching_res = Mock(method='post', path='/stories')
        resource = Mock(root=Mock(resources=[
            matching_res,
            Mock(method='options', path='/stories'),
            Mock(method='post', path='/items'),
        ]))
        mock_get.return_value = None
        models.prepare_relationship('stories', 'Story', resource)
        mock_set.assert_called_once_with(matching_res, 'Story')

    @patch('ramses.models.get_existing_model')
    def test_setup_data_model_existing_model(self, mock_get):
        from ramses import models
        mock_get.return_value = 1
        model, auth_model = models.setup_data_model('foo', 'Bar')
        assert not auth_model
        assert model == 1
        mock_get.assert_called_once_with('Bar')

    @patch('ramses.models.resource_schema')
    @patch('ramses.models.get_existing_model')
    def test_setup_data_model_no_schema(self, mock_get, mock_schema):
        from ramses import models
        mock_get.return_value = None
        mock_schema.return_value = None
        with pytest.raises(Exception) as ex:
            models.setup_data_model('foo', 'Bar')
        assert str(ex.value) == 'Missing schema for model `Bar`'
        mock_get.assert_called_once_with('Bar')
        mock_schema.assert_called_once_with('foo')

    @patch('ramses.models.resource_schema')
    @patch('ramses.models.generate_model_cls')
    @patch('ramses.models.get_existing_model')
    def test_setup_data_model_success(self, mock_get, mock_gen, mock_schema):
        from ramses import models
        mock_get.return_value = None
        mock_schema.return_value = {'field1': 'val1'}
        model = models.setup_data_model('foo', 'Bar')
        mock_get.assert_called_once_with('Bar')
        mock_schema.assert_called_once_with('foo')
        mock_gen.assert_called_once_with(
            schema={'field1': 'val1'},
            model_name='Bar',
            raml_resource='foo')
        assert model == mock_gen()

    @patch('ramses.models.setup_data_model')
    def test_handle_model_generation_value_err(self, mock_set):
        from ramses import models
        mock_set.side_effect = ValueError('strange error')
        with pytest.raises(ValueError) as ex:
            models.handle_model_generation('foo', '/stories')
        assert str(ex.value) == 'Story: strange error'
        mock_set.assert_called_once_with('foo', 'Story')

    @patch('ramses.models.setup_data_model')
    def test_handle_model_generation(self, mock_set):
        from ramses import models
        mock_set.return_value = ('Foo1', True)
        model, auth_model = models.handle_model_generation(
            'foo', '/stories')
        mock_set.assert_called_once_with('foo', 'Story')
        assert model == 'Foo1'
        assert auth_model


@patch('ramses.models.registry')
@pytest.mark.usefixtures('engine_mock')
class TestGenerateModelCls(object):

    def _test_schema(self):
        return {
            'properties': {},
            '_auth_model': False,
            '_public_fields': ['public_field1'],
            '_auth_fields': ['auth_field1'],
            '_nested_relationships': ['nested_field1']
        }

    @patch('ramses.models.resolve_to_callable')
    def test_simple_case(self, mock_res, mock_reg):
        from nefertari.authentication.models import AuthModelMethodsMixin
        from ramses import models
        models.engine.FloatField.reset_mock()
        schema = self._test_schema()
        schema['properties']['progress'] = {
            "_db_settings": {
                "type": "float",
                "required": True,
                "default": 0,
                "before_validation": ["zoo"],
                "after_validation": ["foo"],
                "backref_after_validation": ["foo"],
                "backref_before_validation": ["foo"]
            }
        }
        mock_res.return_value = 1
        mock_reg.mget.return_value = {'foo': 'bar'}
        model_cls, auth_model = models.generate_model_cls(
            schema=schema, model_name='Story', raml_resource=None)
        assert not auth_model
        assert model_cls.__name__ == 'Story'
        assert hasattr(model_cls, 'progress')
        assert model_cls.__tablename__ == 'story'
        assert model_cls._public_fields == ['public_field1']
        assert model_cls._auth_fields == ['auth_field1']
        assert model_cls._nested_relationships == ['nested_field1']
        assert model_cls.foo == 'bar'
        assert issubclass(model_cls, models.engine.ESBaseDocument)
        assert not issubclass(model_cls, AuthModelMethodsMixin)
        models.engine.FloatField.assert_called_once_with(
            default=0, required=True, before_validation=[1],
            after_validation=[1], backref_after_validation=[1],
            backref_before_validation=[1])
        mock_res.assert_has_calls([call('zoo'), call('foo')])
        mock_reg.mget.assert_called_once_with('Story')

    @patch('ramses.models.resolve_to_callable')
    def test_callable_default(self, mock_res, mock_reg):
        from ramses import models
        models.engine.FloatField.reset_mock()
        schema = self._test_schema()
        schema['properties']['progress'] = {
            "_db_settings": {
                "type": "float",
                "default": "{{foobar}}",
            }
        }
        mock_res.return_value = 1
        model_cls, auth_model = models.generate_model_cls(
            schema=schema, model_name='Story', raml_resource=None)
        models.engine.FloatField.assert_called_with(
            default=1, required=False)
        mock_res.assert_called_once_with('{{foobar}}')

    def test_auth_model(self, mock_reg):
        from nefertari.authentication.models import AuthModelMethodsMixin
        from ramses import models
        schema = self._test_schema()
        schema['properties']['progress'] = {'_db_settings': {}}
        schema['_auth_model'] = True
        mock_reg.mget.return_value = {'foo': 'bar'}

        model_cls, auth_model = models.generate_model_cls(
            schema=schema, model_name='Story', raml_resource=None)
        assert auth_model
        assert issubclass(model_cls, AuthModelMethodsMixin)

    def test_db_based_model(self, mock_reg):
        from nefertari.authentication.models import AuthModelMethodsMixin
        from ramses import models
        schema = self._test_schema()
        schema['properties']['progress'] = {'_db_settings': {}}
        mock_reg.mget.return_value = {'foo': 'bar'}

        model_cls, auth_model = models.generate_model_cls(
            schema=schema, model_name='Story', raml_resource=None,
            es_based=False)
        assert issubclass(model_cls, models.engine.BaseDocument)
        assert not issubclass(model_cls, models.engine.ESBaseDocument)
        assert not issubclass(model_cls, AuthModelMethodsMixin)

    def test_no_db_settings(self, mock_reg):
        from ramses import models
        schema = self._test_schema()
        schema['properties']['progress'] = {'type': 'pickle'}
        mock_reg.mget.return_value = {'foo': 'bar'}

        model_cls, auth_model = models.generate_model_cls(
            schema=schema, model_name='Story', raml_resource=None,
            es_based=False)
        assert not models.engine.PickleField.called

    def test_unknown_field_type(self, mock_reg):
        from ramses import models
        schema = self._test_schema()
        schema['properties']['progress'] = {
            '_db_settings': {'type': 'foobar'}}
        mock_reg.mget.return_value = {'foo': 'bar'}

        with pytest.raises(ValueError) as ex:
            models.generate_model_cls(
                schema=schema, model_name='Story', raml_resource=None)
        assert str(ex.value) == 'Unknown type: foobar'

    @patch('ramses.models.prepare_relationship')
    def test_relationship_field(self, mock_prep, mock_reg):
        from ramses import models
        schema = self._test_schema()
        schema['properties']['progress'] = {
            '_db_settings': {
                'type': 'relationship',
                'document': 'FooBar',
            }
        }
        mock_reg.mget.return_value = {'foo': 'bar'}
        models.generate_model_cls(
            schema=schema, model_name='Story', raml_resource=1)
        mock_prep.assert_called_once_with('progress', 'FooBar', 1)

    def test_foreignkey_field(self, mock_reg):
        from ramses import models
        schema = self._test_schema()
        schema['properties']['progress'] = {
            "_db_settings": {
                "type": "foreign_key",
                "ref_column_type": "string"
            }
        }
        mock_reg.mget.return_value = {'foo': 'bar'}
        models.generate_model_cls(
            schema=schema, model_name='Story', raml_resource=1)
        models.engine.ForeignKeyField.assert_called_once_with(
            required=False, ref_column_type=models.engine.StringField)

    def test_list_field(self, mock_reg):
        from ramses import models
        schema = self._test_schema()
        schema['properties']['progress'] = {
            "_db_settings": {
                "type": "list",
                "item_type": "integer"
            }
        }
        mock_reg.mget.return_value = {'foo': 'bar'}
        models.generate_model_cls(
            schema=schema, model_name='Story', raml_resource=1)
        models.engine.ListField.assert_called_once_with(
            required=False, item_type=models.engine.IntegerField)

    def test_duplicate_field_name(self, mock_reg):
        from ramses import models
        schema = self._test_schema()
        schema['properties']['_public_fields'] = {
            '_db_settings': {'type': 'interval'}}
        mock_reg.mget.return_value = {'foo': 'bar'}
        models.generate_model_cls(
            schema=schema, model_name='Story', raml_resource=1)
        assert not models.engine.IntervalField.called
